#include <stdio.h>
#include <string.h>	// for memset
#include "Globals.h"
#include "gpib.h"
#include "db_utils.h"
#include "lifdisk.h"
#include "amigo.h"
#include "ff.h"
#include "xprintf.h"
#include "sio.h"    // For xputch(), remove after debugging)
//
// Amigo disk handling. See "Subset 80 from Fixed and flexible disc drives" 5958-4129 and the 9123 disc drive manual addendum 5957-6584
//
// HP 9121 Amigo disk is stored on the SD card as a disk image cylinder - head - sector
//
// Docs:
// LIF filesystem http://www.hp9845.net/9845/projects/hpdir/#lif_filesystem
// Commands: Manual Update 5957-6584 for the 9123D
// Appendix A: HP 9895A Disc Memory Command Set

//extern char string[MAX_DBG_STRING];

// ----------------------------------------- Structures used for testing ------------------------------------------
// LIF data is BIG endian
//
// Volume label structure
//
VolumeLabelType vl = {0x0080,{'L','A','B','E','L','1'},
0,0x200,
0x0080,
0,
0,0x0e00,						// Change this to 0x0e -> LIF directory too big
0
};
//
// Directory entry
//
DirEntryType de = {{'M','Y','F','I','L','E','1','2','3','4'},0x100,0,0x100,0,0x20,{0x13,0x10,0x12,0x00,0x00,0x00},1,0,0};
//
// Structure to hold head/cylinder/sector
//
DiskType Disk = {0,0,0,0};
//--------------------------------------------------------------------------------------------------------------------
unsigned char AM_talking;
unsigned char AM_listening;
unsigned volatile int amigo_state;
unsigned char dsj;
unsigned char amigo_buff_flag = AMIGO_BUFFERED;      // Shows if 0x60 should be buffered or unbuffered
unsigned char amigo_error;
//
// Init Amigo state machine
//
void amigo_init()
{
	AM_talking = 0;
	AM_listening = 0;
	amigo_state = AMIGO_START;
	dsj = 2;
	EnablePP();
    amigo_buff_flag = AMIGO_BUFFERED;
    amigo_error = 0;
}
//
// This is the command parser. Parses byte by byte from gpib
//
int amigo_char(unsigned char ch, unsigned char atn, unsigned char reset)
{
unsigned char status;

	if(reset) amigo_init();
	status = 0;
	//
//	if(AM_listening)
//		xprintf("{L%u %d}[%02X]", amigo_state,atn,ch);
//	else if(AM_talking)
//		xprintf("{T%u %d}[%02X]", amigo_state,atn,ch);
//	else
//		xprintf("{%u %d}[%02X]", amigo_state, atn,ch);
    //xprintf("(%X %d)", ch, atn);
	//
    xprintf("[%X %d]", ch, PP_LATCH);
	// Handle single-byte commands, they all reset the state-machine
	//
	if(atn)
	{
        // Strip parity bit
        //
        ch = ch & 0x7F;
        //sio_putch('a');
		if(ch == SDC || ch == DCL)						// Selected device clear or all device clear
		{
			status = AMIGO_CMD_CLEAR;
			amigo_state = AMIGO_START;
			return status;
		}
        //sio_putch('b');
		if(ch == UNL)
		{
			AM_listening = 0;
			amigo_state = AMIGO_START;
			return 0;
		}
        //sio_putch('c');
		if(ch == UNT)
		{
			AM_talking = 0;
			amigo_state = AMIGO_START;
			return 0;
		}
        //sio_putch('d');
		if(ch >= 0x40 && ch <= 0x5e && ch != MTA)		// A talk address automatically untalks everyone
		{
			AM_talking = 0;
			amigo_state = AMIGO_START;
			return 0;
		}
		if(ch == MTA)
		{
			AM_talking = 1;
            AM_listening = 0;
			amigo_state = AMIGO_START;
#if SDEBUG > 1
			xprintf("[MTA]");
#endif
			return 0;
		}
        //sio_putch('e');
		if(ch == MLA)
		{
			AM_listening = 1;
			amigo_state = AMIGO_START;
#if SDEBUG > 1
			xprintf("[MLA]");
#endif
			return 0;
		}
	}
	//
	//xprintf("\r\nastate %d, ch %X", amigo_state, ch);
	//

	switch(amigo_state)
	{
		case AMIGO_START:
			//
			if(ch == 0x60 && AM_talking)		// Single byte command
			{                                   // Execute read
				//xprintf("\r\nAM:Buffering flag %d", amigo_buff_flag);
                if(amigo_buff_flag)
                    status = AMIGO_CMD_UNBUFFERED_READ;
                else
                    status = AMIGO_CMD_BUFFERED_READ;
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x60 && AM_listening)		// Single byte command
			{                                   // execute write
				//xprintf("\r\nBuffered write", Disk.sector);
                if(amigo_buff_flag)
                    status = AMIGO_CMD_UNBUFFERED_WRITE;
                else
                    status = AMIGO_CMD_BUFFERED_WRITE;
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x68 && AM_listening)
            {
                amigo_state = AMIGO_068_L;
                DisablePP();
                break;
            }
			if(ch == 0x68 && AM_talking)		// Single byte cmd
			{
				//xprintf("\r\nSend status", Disk.sector);
				status = AMIGO_CMD_SEND_STATUS;
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x6A && AM_listening)
            {
                amigo_state = AMIGO_06A;
                DisablePP();
                break;
            }
			if(ch == 0x6C && AM_listening)  // Format
            {
                amigo_state = AMIGO_06C_L;
                DisablePP();
                break;
            }
			if(ch == 0x70 && AM_talking) 	// DSJ is a single byte command
			{
				//xprintf("\r\nDSJ");
				status = AMIGO_CMD_DSJ;
				amigo_state = AMIGO_START;
				DisablePP();
				break;
			}
            if(ch == 0x70 && AM_listening)
            {
                amigo_state = AMIGO_070_L;
                DisablePP();
                //trigger(4);
            }
            break;

		case AMIGO_068_L:
			if(ch == 0x02) amigo_state = AMIGO_068_SEEK;
            if(ch == 0x05) amigo_state = AMIGO_068_READ;
            if(ch == 0x07) amigo_state = AMIGO_068_VERIFY;
			if((ch == 0x03) || (ch == 0xff))
			{
                //trigger(3);
				DisablePP();
                // Todo
                status = 0; // Do nothing just a request wait for talk. Enable PP and wait for talk, then the 0x68 cmd
				EnablePP();

				//status = AMIGO_CMD_SEND_STATUS;
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x16) amigo_state = AMIGO_068_WAKEUP;
			break;
		case AMIGO_068_T:
			//xprintf("\r\nSend status", Disk.sector);
			status = AMIGO_CMD_SEND_STATUS;
			amigo_state = AMIGO_START;
			break;
		case AMIGO_06A:
			if(ch == 0x05) amigo_state = AMIGO_06A_READ;
			break;
		case AMIGO_068_SEEK:					// seek, five bytes follow unknown, unit, head,sector
			amigo_state = AMIGO_068_SEEK_1;
			break;
		case AMIGO_068_SEEK_1:
			Disk.unit = ch;
			amigo_state = AMIGO_068_SEEK_2;
			break;
		case AMIGO_068_SEEK_2:
			Disk.cyl = ch;
			amigo_state = AMIGO_068_SEEK_3;
			break;
		case AMIGO_068_SEEK_3:
			Disk.head = ch;
			amigo_state = AMIGO_068_SEEK_4;
			break;
		case AMIGO_068_SEEK_4:
			Disk.sector = ch;
			//xprintf("\r\nSeek: %d %d %d %d, %X", Disk.unit, Disk.cyl, Disk.head, Disk.sector, ch);
			amigo_state = AMIGO_START;
			status = AMIGO_CMD_SEEK;
			break;
			//
		case AMIGO_068_VERIFY:					// seek, three bytes follow  unit, 2 bytes sector count
			amigo_state = AMIGO_068_VERIFY_1;
            DisablePP();
			break;
        case AMIGO_068_VERIFY_1:				// Unit byte
			amigo_state = AMIGO_068_VERIFY_2;
			break;
        case AMIGO_068_VERIFY_2:				// sector count hign
			amigo_state = AMIGO_068_VERIFY_3;
			break;
        case AMIGO_068_VERIFY_3:				// sector count low
			amigo_state = AMIGO_START;
            xprintf("** Verify **");
            //__delay_ms(1000);
            EnablePP()
			break;
		case AMIGO_068_WAKEUP:
			amigo_state = AMIGO_START;
			status = AMIGO_CMD_WAKEUP;
			break;							
		case AMIGO_06A_READ:							// Does nothing, but prepares for the read. Actual read bytes is AMIGO_60
            amigo_buff_flag = AMIGO_BUFFERED;           // Set flag to show read type
            amigo_state = AMIGO_06A_READ_2;				//two bytes follow <- Wrong, one byte, skip to read2
            // Todo
			//amigo_state = AMIGO_06A_READ_1;			//This was the drive byte
            trigger(2);
            EnablePP();                                 // enable PP to start read
			break;
		case AMIGO_06A_READ_1:							// Read byte 1
			amigo_state = AMIGO_06A_READ_2;
			break;
		case AMIGO_06A_READ_2:							// Read byte 2 (unit))
            EnablePP();                                 // Enable parallell poll
			amigo_state = AMIGO_END;
            trigger(3);
			break;
		case AMIGO_06C_L:                               
			if(ch == 0x18) amigo_state = AMIGO_06C_FORMAT;  // Format. Four bytes follow
            break;
        case AMIGO_06C_FORMAT:                              // This byte is unit
            amigo_state = AMIGO_06C_FORMAT1;
            break;
        case AMIGO_06C_FORMAT1:                             // This byte is type, 2=HP, 8= IBM
            amigo_state = AMIGO_06C_FORMAT2;
            break;
        case AMIGO_06C_FORMAT2:                             // This byte is interleave
            amigo_state = AMIGO_06C_FORMAT3;
            break;
        case AMIGO_06C_FORMAT3:                             // This byte is an unspecified data byte
			amigo_state = AMIGO_START;
			status = AMIGO_CMD_FORMAT;
            EnablePP();
            break;
		case AMIGO_070_T:									// This is DSJ, put status byte on bus
			//xprintf("\r\nDSJ");
			//amigo_state = AMIGO_START;
			status = AMIGO_CMD_DSJ;
			amigo_state = AMIGO_END;
            DisablePP();                                    // DSJ leaves PP disabled
			break;
		case AMIGO_070_L:									// This is Amigo clear. Seek to 0/0/0 and put status byte on bus
            DisablePP();
            //trigger(5);
            amigo_state = AMIGO_START;
            status = 0;
            break;
        case AMIGO_068_READ:                                // Disk in ch, unbuffered read. Need to set a flag
            EnablePP();                                     // ao that the execution command does the right
            //trigger(6);                                     // type of read
            amigo_buff_flag = AMIGO_UNBUFFERED;
            status = AMIGO_CMD_UNBUFFERED_READ;
            amigo_state = AMIGO_START;
            status = 0;
            break;            
        //case AMIGO_070_L_1:
            //amigo_state = AMIGO_END;
            
        case AMIGO_END:
		default:
			break;
	}
	//xprintf("(%d)", status);
	return status;
}
//
// Return drive status to host. TODO: Fix stuff like proper status for missing SD etc
//
// Status bytes
// 1: 00DSSSSS , S=0 => normal completion
// 2: 0000UUUU , Unit
// 3: 00*TTTTR , * is set drive fault, seek check, not ready, TTTT=0110->HP, R=1 -> 9121
// 4: AWxEFCSS , Atten, WriteProt,x,Fault, First status, seek check, status
//
void amigo_cmd_send_status()
{
#if SDEBUG >= 1
	xprintf("\r\nAM:Send status");
#endif
	//
	gpib_flags = gpib_flags | FLAGS_EOI;
	gpib_tx_buff[0] = amigo_error;		// Status 1
	gpib_tx_buff[1] = 0x00;				//
	gpib_tx_buff[2] = 0x0d;				// Status 2 0110 = hp format, 1=HP9121
	gpib_tx_buff[3] = 0x00;				//
	if(WP_PORT) gpib_tx_buff[3] = 0x60;	// Write protect
    if(WP_PORT) gpib_tx_buff[2] = gpib_tx_buff[2] | 0b00100000;	// Status 2 error
	sendgpib(4);
	gpib_flags = gpib_flags & ~FLAGS_EOI;
    amigo_error = 0;
	EnablePP();
}
//
// Seek to the position held in the disk structure
// 9121 16 sectors/track
// 9895 30 sectors/track
//
void amigo_cmd_seek()
{
DWORD fpos;
FRESULT rc;
	fpos = (DWORD)Disk.cyl*16L*256L*2L + (DWORD)Disk.head * 256L *16L + (DWORD)Disk.sector *256L;
#if SDEBUG >= 1
	xprintf("\r\nAM: Seek: %d %d %d %d, s:%ld", 
		Disk.unit, Disk.cyl, Disk.head, Disk.sector,
		(DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector
		);
	xprintf("\r\nSeeking to %ld (%lX)", fpos, fpos);
#endif
	rc=f_lseek(&File[0],fpos);
	if(rc)
	{
		xprintf("\r\nSeek to %ld failed %d", fpos, rc);
	}
	dsj = 0;
	EnablePP();
}
//
// Do a buffered read of the current sector
//
void amigo_cmd_buffered_read()
{
DWORD total_bytes;
FRESULT rc;
UINT nbytes;
UINT sbytes;
						// Follows after 0x680x05, Send  256 bytes. EOI only if error
#if SDEBUG >= 1
	xprintf("\r\nAM:Buff rd c/h/s: %d/%d/%d s:%ld ", Disk.cyl, Disk.head, Disk.sector,
		(DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector
	);
#endif
	//
	total_bytes = 0;
	//for(i=0;i<4;i++)	// Read 256 bytes
	//{
		rc=f_read(&File[0],gpib_tx_buff,256,&nbytes);
#if SDEBUG >= 1
		xprintf("\r\nRead returns %X, %d bytes read", rc, nbytes);
#endif

		gpib_flags = gpib_flags & ~FLAGS_EOI;
		//gpib_tx_buff[0] = ' ';
		//gpib_tx_buff[1] = nbytes+2;
        can_break = 1;
		sbytes=sendgpib(nbytes);
        can_break = 0;
		total_bytes = total_bytes + sbytes;
	//}
#if SDEBUG >= 1
	xprintf("\r\n %ld bytes sent", total_bytes);
#endif
	dsj = 0;
	EnablePP();	
}
//
// Write 256 bytes at current position, then advance to next sector
//
//
void amigo_cmd_buffered_write()
{
FRESULT rc;
UINT nbytes;
int gpib_bytes;
char b,c;

						// Follows after 0x690x08, receive  max 256 bytes or up to EOI
#if SDEBUG >= 1
	xprintf("\r\nAM:Buff wr c/h/s: %d/%d/%d s:%ld", Disk.cyl, Disk.head, Disk.sector,
		(DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector
	);
#endif
    
    gpib_bytes = 0;
    nbytes = 0;
    c = 0;
    //xprintf("\r\nEntry, gpib_bytes %d ", gpib_bytes);
    while(gpib_bytes < 256)
    {
        if(gpib_test())             // We have a character from GPIB
        {
            b = gpib_getc(&c);
            //xprintf(" %X %X %d ", b, c, gpib_bytes);
            gpib_rx_buff[gpib_ptr++] = b;
            ++gpib_bytes;
#if SDEBUG >= 9
            xprintf("\r\n %02X %c %4s %4s %4s %4s %4s %d",
            b,b,
            c&IFC_BIT ? "IFC ":"  ",
            c&SRQ_BIT ? "SRQ ":"  ",
            c&ATN_BIT ? "ATN ":"  ",
            c&EOI_BIT ? "EOI ":"  ",
            c&REN_BIT ? "REN ":"  ",
                    gpib_ptr);
#endif
            //if(c&EOI_BIT) break;
        // IFC
        }
        if(c&EOI_BIT)
        {
            //xprintf(" **EOI**");
             break;
        }
        if(IFC_PORT == 0) break;
    }

    nbytes = gpib_bytes;
	//
    if(!WP_PORT) rc=f_write(&File[0],(void*)gpib_rx_buff, 256, &nbytes);
	if(rc)
	{
		xprintf("\r\nWrite fail %d", rc);
	}
    
#if SDEBUG >= 1
	xprintf("\r\nAM:Buff wr %d bytes rx", nbytes);
#endif
	amigo_increment();			// The address is incremented on a successful read.
	dsj = 0;
	EnablePP();
}
//
// DSJ, return status byte
//
void amigo_cmd_dsj()
{
#if SDEBUG >= 1
	xprintf("\r\nAM:DSJ");
    //trigger(4);
#endif
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	gpib_tx_buff[0] = dsj;
	sendgpib(1);	
	//
	// DSJ clears the DSJ byte that might be 2 after powerup
	//
	dsj = 0;
}

//
// Wakeup (undocumented), just enable PP response
//
//  54 3F 54 20 68 16 00 
//
void amigo_cmd_wakeup()
{
#if SDEBUG >= 1
	xprintf("\r\nAM:Wakeup");
#endif
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	//gpib_tx_buff[0] = ' ';
	//gpib_tx_buff[1] = 3;	// 2+1 byte
	//gpib_tx_buff[2] = 0;	// Status OK
	gpib_tx_buff[0] = dsj;
	//sendgpib(1);
	EnablePP();
	//
	// Wakeup clears the DSJ byte that might be 2 after powerup
	//
	dsj = 0;
}
//
// Response to selected device clear or all clear - Select sector 0
//
//
void amigo_cmd_clear()
{
#if SDEBUG >= 1
	xprintf("\r\nAM:Clear");
#endif
	Disk.sector = 0;
	Disk.head = 0;
	Disk.cyl = 0;
	EnablePP();
	//
	// Clear the DSJ byte that might be 2 after powerup
	//
	dsj = 0;
}
//
// Increment sector (and possibly head and cylinder) then seek to that position
//
void amigo_increment()
{
DWORD fpos;
FRESULT rc;

	++Disk.sector;
	if(Disk.sector > 15)
	{
		Disk.sector = 0;
		++Disk.head;
		if(Disk.head > 1)
		{
			Disk.head = 0;
			++Disk.cyl;
		}
	}
	fpos = (DWORD)Disk.cyl*16L*256L*2L + (DWORD)Disk.head * 256L *16L + (DWORD)Disk.sector *256L;
#if SDEBUG >= 1
	xprintf("\r\nAM: Incr: %d %d %d %d, ", Disk.unit, Disk.cyl, Disk.head, Disk.sector);
	//xprintf("\r\nSeeking to %ld (%lX)", fpos, fpos);
#endif
	rc=f_lseek(&File[0],fpos);
	if(rc)
	{
		xprintf("\r\nSeek to %ld failed %d", fpos, rc);
	}
	dsj = 0;
}
//
// Do an unbuffered read of the current sector. Send current sector
// Then increment and send next until the host says STOP by 
// asserting ATN and sending a command
//
void amigo_cmd_unbuffered_read()
{
DWORD total_bytes;
FRESULT rc;
UINT nbytes;
UINT sbytes;
int more = 1;
int i;
						//  EOI only if error
#if SDEBUG >= 1
	xprintf("\r\nAM:Unbuff rd c/h/s: %d/%d/%d s:%ld ", Disk.cyl, Disk.head, Disk.sector,
		(DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector
	);
#endif
	//
	total_bytes = 0;
	while(more)	
	{
		rc=f_read(&File[0],gpib_tx_buff,256,&nbytes); // Read 256 bytes
#if SDEBUG >= 1
		//xprintf("\r\nRead returns %X, %d bytes read\r\n", rc, nbytes);
        xprintf("\r\nrc %X, %d bytes\r\n", rc, nbytes);
#endif

		gpib_flags = gpib_flags & ~FLAGS_EOI;
        //can_break = 0;
		sbytes = sendgpib(nbytes);
        //can_break = 0;
        //xprintf(" sbytes %d ", sbytes);
        //if(Disk.sector == 3 && Disk.cyl == 32) 
        //trigger(4);
		total_bytes = total_bytes + sbytes;
        if (IFC_PORT == 0)
        {
            //trigger(5);
            break;
        }
        // Did we send less than read from disk? If so, the host has
        // stopped transmission by asserting ATN
        //
        if(sbytes != nbytes)
        {
            more = 0;
                //trigger(5);
            xprintf("\r\nsbytes %d", sbytes);
            break;
        }
//        __delay_ms(100);
//        if(!talking)
//            {
//                more = 0;
//                xprintf("\r\nUntalked ");
//                break;
//            }
        gpib_flush();
        if(gpib_test())
        {
                more = 0;
                xprintf("\r\nHost bytes %d", gpib_test());
                dsj = 1;
                break;
        }
        //
        // Do we really need the increment? We are at the correct
        // position already??
        //amigo_increment();
        // Check for termination from host
        i=5;
        while(i--)
        {
            __delay_ms(10);
            if(gpib_test())
            {
                more = 0;
                //trigger(5);
                xprintf("\r\nHost byte2 %d", i);
                break;
            }
            //dump();
            //xprintf("\r\nTalking %d", talking);
        }
        //    dump();
        //trigger(5);
//        if(ATN_PORT == 0)
//        {
//            xprintf(" Got ATN");
//            more = 0;
//        }
        if(gpib_test()) more = 0;
	} // more
#if SDEBUG >= 1
	xprintf("\r\n %ld bytes sent", total_bytes);
#endif
	dsj = 0;
	EnablePP();	
}
void amigo_cmd_unbuffered_write()
{
}

// Format a disk. Rename the old lifdata file and create a new. Write all blocks
//
void amigo_cmd_format()
{
UINT i;
UINT nbytes;
BYTE res;
char b,c;
int status;

    DisablePP();
    memset((void*)gpib_rx_buff, 0xF6,256);
    xprintf( "\r\nFormatting ");
    //xprintf("\r\nWrite %ld (%lX) s:%ld ", Vector, Vector,sector);
    for(i=0;i<200;i++)
    //
	{

        //xprintf("blk %05ul ",i);
        sio_putch('.');
		if(!WP_PORT) res=f_write(&File[0],(void*)gpib_rx_buff, 256, &nbytes);
        if(res != FR_OK)
        {
            xprintf("rc=%d ",res);
            return;
        }
        if(gpib_test())     // We have a character from GPIB
        {
            b = gpib_getc(&c);
            status = amigo_char(b,c&ATN_BIT,0);

            if(status == AMIGO_CMD_DSJ) amigo_cmd_dsj();
            if(status == AMIGO_CMD_SEND_STATUS) amigo_cmd_send_status();
            if(status == AMIGO_CMD_SEEK) amigo_cmd_seek();
        }
	}
    xprintf( "Fmt exit ");
	EnablePP();
}