// Based on code from Raphael Kuhnen <raphael.kuhnen@fmf.uni-freiburg.de> and others.
//
// (c) 2010 - 2017 Anders Gustafsson <anders.gustafsson@pedago.fi> 
//
// Credits: Ruud Baltissen - http://ftp.osuosl.org/pub/nslu2/sources/xmame-0.100.tar.bz2
//
// PIC Pin addignments:
//
// RA0
// RA1 - IFC
// RA2 -      SRQ, moved to A7
// RA3 - ATN
// RA4 - EOI
// RA5 - REN
// RA6 - 
// RA7 - SRQ

//
// RB0 - D1
// RB1 - D2
// RB2 - D3
// RB3 - D4
// RB4 - D5
// RB5 - D6
// RB6 - D7
// RB7 - D8
//
// RC0 - DAV
// RC1 - NRFD
// RC2 - NDAC
// RC3 - na
// RC4 - USB D-
// RC5 - USB D+
// RC6 - TX
// RC7 - RX
//

//
// GPIB Connector pinout
//
//Pin  	Name 	Signal Description 			Pin Name 	Signal Description 
//1 	DIO1 	Data Input/Output Bit 1 	13 	DIO5 	Data Input/Output Bit 5 
//2 	DIO2 	Data Input/Output Bit 2 	14 	DIO6 	Data Input/Output Bit 6 
//3 	DIO3 	Data Input/Output Bit 3 	15 	DIO7 	Data Input/Output Bit 7 
//4 	DIO4 	Data Input/Output Bit 4 	16 	DIO8 	Data Input/Output Bit 8 
//5 	EIO 	End-Or-Identify 			17 	REN 	Remote Enable 
//6 	DAV 	Data Valid 					18 	Shield 	Ground (DAV) 
//7 	NRFD 	Not Ready For Data 			19 	Shield 	Ground (NRFD) 
//8 	NDAC 	Not Data Accepted 			20 	Shield 	Ground (NDAC) 
//9 	IFC 	Interface Clear 			21 	Shield 	Ground (IFC) 
//10 	SRQ 	Service Request 			22 	Shield 	Ground (SRQ) 
//11 	ATN 	Attention 					23 	Shield 	Ground (ATN) 
//12 	Shield 	Chassis Ground 				24 	Single GND 	Single Ground

#include <xc.h>
#include <stdio.h>
#include <string.h>
#include "gpib.h"
//#include "prologix_cmd.h"
#include "db_utils.h"
#include "sio.h"
#include "xprintf.h"
#include "integer.h"
#include "defines.h"

// Circular buffer (FiFo) Needs to be larger than secttor size, plus space
// To accomodate preceeding command bytes. 384 seemed to work so 512 gives some 
// Margin
//
// Handshake is done with NRFD. If buffer fills up, then NRFD is asserted.
// NRFD is then de-asserted when you remove bytes from the buffer
//
#define BUFFER_SIZE 512
static volatile struct {
	int		ri, wi, ct;			/* Read index, Write index, Data counter */
	BYTE	buff[BUFFER_SIZE];	/* FIFO buffer, received GPIB */
    BYTE	cbuff[BUFFER_SIZE]; // control bytes (ATN and stuff)
} RxFifo;

char gpib_rx_buff[GPIB_MAX_RX_LEN];
char gpib_tx_buff[512]; //GPIB_MAX_TX_LEN];
//
//
extern volatile BYTE gpib_debug;
//
// Put often used vars in access ram
unsigned char eoiflag;
unsigned char atnflag;

unsigned char mla;
unsigned char mta;
unsigned char spoll;
unsigned char spbyte;

char gpib_ptr;
//
unsigned int gpib_tx_ptr;
int i;
int j;
unsigned long count, timeout;

unsigned char rx_state;				// Rx state machine state
unsigned char tx_state;				// Tx state machine state

unsigned char gpib_addr = 0;

unsigned char gpib_flags;
char gpib_eot_char;
unsigned char can_break;
//
// Set sensible initial values
//
void gpib_init()
{
    DIO_LAT = 0;                // Initial values, data and status inputs
    STATUS_LAT = 0;
    DIO_TRIS = 0xFF;
    STATUS_TRIS = 0xFF;
    can_break = 0;
	//
 	// Zero the GPIB buffers and pointers
	//
	gpib_ptr = 0;
	gpib_tx_ptr = 0;
	listening = 0;
	talking = 0;
	spoll = 0;
	spbyte = 0;

	//
	mla = gpib_addr + (unsigned) MLA;
	mta = gpib_addr + (unsigned) MTA;
	i=0;
	while(i<GPIB_MAX_RX_LEN)
	{
		gpib_rx_buff[i] = 0;
		i++;
	}
	i=0;
	while(i<GPIB_MAX_TX_LEN)
	{
		gpib_tx_buff[i] = 0;
		i++;
	}
    // Open Drain on status port
    ODCONB = 0b11111111;
    //
    // Timer 0 as timeout counter. 64mHz / 4 / 16 / 16 = 16us, timeout 4ms
    T0CON1bits.T0CS = 2;            // Timer 0 source Fosc/4
    T0CON1bits.T0ASYNC = 0;
    T0CON1bits.T0CKPS = 0x0f;       // divide by 16 prescaler
    T0CON0bits.T016BIT = 0;         // Timer 0 as 8-bit counter
    T0CON0bits.T0OUTPS = 0x0f;      // divide by 16 postscaler
    PIE0bits.TMR0IE = 0;            // No timer interrupt
    T0CON0bits.T0EN = 1;            // Enable timer
}

//-----------------------------------------------------------------------------------------------------------------------
// Send routines are written as finite-state-machines. Please refer to the timing diagram on page 14 of
// "A tutorial Description of the Hewlett-Packard Interface Bus" (http://bitsavers.org/pdf/hp/hpib/TutorialDescrOfHPIB.pdf)
// Reception is now an ISR
//-----------------------------------------------------------------------------------------------------------------------
//
// sendgpib(length) - Send GPIB string in gpib_tx_buff[]
//
//
// Assumption is that system is set as listener when we arrive here
int sendgpib(int len)
{
	tx_state = GPIB_TX_START;
	j=0;
	REN_TRIS = 1;

	DAV_TRIS = 1;
	NRFD_TRIS = 1;
	NDAC_TRIS = 1;
	ATN_TRIS = 1;
	DIO_TRIS = 0xff;
    DAV_LAT = 0;
	NRFD_LAT = 0;
	NDAC_LAT = 0;
	ATN_LAT = 0;
    //
    LATAbits.LATA7 = 1; // Disk LED
	//
	// We should not talk unless given permission
	//
	if(!talking)
	{
#if defined(SDEBUG)
		if(gpib_debug & GPIB_DBG_TX)
		{
			xprintf("\r\nTried to talk whilst not addressed"); // 
		}
#endif
		return 0;
	}
    //
    //PIE0bits.IOCIE = 0;     // Disable interrupt-on-change
    IOCBNbits.IOCBN0 = 0;   // disable Interrupt on negative-going edge DAV
    __delay_us(10);
    //
    //xprintf("\r\nTXDATA len %d ", len);
//    while(ATN_PORT == 0 )
//    {
//        xprintf(".");
//    }
#if SDEBUG > 1
//	if(atn == 1) 				// Is it a command?
//	{
//		xprintf("\r\nTXCMD len: %d ",len); // 
//	} else
//	{
		xprintf("\r\nTXDATA len %d ", len);
	//}
	DumpData(gpib_tx_buff, len, "TXData");
#endif
	//
	while(tx_state != GPIB_TX_DONE && IFC_PORT == 1)
	{
		//xprintf("\r\nTX State:%d  %X %X       ", tx_state ,DIO_PORT, STATUS_PORT);
		//sio_puts(string);

#if SDEBUG >= 1
		if(ATN_PORT == 0 && EOI_PORT == 0)
		{
			xprintf("{PP%d %d %d}",PP_PORT, PP_LATCH, tx_state);	// 
            //tx_state = GPIB_TX_DONE;
            //continue;
		}
        
#endif
        if(ATN_PORT == 0)               // We have been interrupted
        {
            //trigger(2);
            //talking = 0;
            //STATUS_TRIS = 0xff;
            DIO_TRIS = 0xff;            // Make sure DIO is input to receive next cmd
            IOCBNbits.IOCBN0 = 1;       // enable Interrupt on negative-going edge DAV
            return j;                   // return number of bytes sent
        }
        //
        if(ATN_PORT == 0)
        {
            xprintf(" ATN asserted ");
            tx_state = GPIB_TX_DONE;
            break;
        }
		count = TIMEOUT;
		//TimeGPIB(tx_state);
		switch(tx_state)
		{
			case GPIB_TX_START:
				if(NRFD_PORT == 1 && NDAC_PORT == 1)		// If both NRFD and NDAC high, then we have an error
				{
					if(count-- <=0)
					{
						tx_state = GPIB_TX_ERROR;
						break;
					}
				}
				else tx_state = GPIB_TX_CHECK;
				break;
			case GPIB_TX_CHECK:
                if(ATN_PORT == 1) tx_state = GPIB_TX_PUT_DATA;
                break;
			case GPIB_TX_PUT_DATA:
                //__delay_us(1);
#if SDEBUG >= 9
				xprintf("<%02X>", gpib_tx_buff[j] & 0xff); // Seems to need a small delay here...
																						// Removing this shaves 1300 ticks
#endif
				if (j == len - 1 && (gpib_flags & FLAGS_EOI)!=0) // Last char and request eoi??
				{
					EOI_TRIS = 0;
#if SDEBUG >= 9
					sio_puts("{EOI}");
#endif
				}
				DIO_LAT = ~gpib_tx_buff[j];                         // Put data on bus
				++j;

				DIO_TRIS = 0; 										// Make data lines output
																	//		A GPIB Talker, transfers data bytes to GPIB Listeners, it drives
																	//		the data byte on the GPIB DIO[8?1] signals. After waiting for a certain delay (known as
																	//		the T1 delay), the Talker asserts DAV to indicate to the Listeners that the data byte
																	//		has settled on the DIO[8?1] signals. A delay of 2us should satisfy most implementations.
																	//		http://www.ni.com/pdf/manuals/370875a.pdf
				tx_state = GPIB_TX_WAIT_FOR_NRFD;
				count = TIMEOUT;
				//TMR0L = 0;                                          // Alternative timeout on TMR0 overflow
				//PIR0bits.TMR0IF = 0;
                //__delay_us(1);
				break;
			case GPIB_TX_WAIT_FOR_NRFD:								// Wait for NRFD high
				if (NRFD_PORT == 1 || count--<=0) tx_state = GPIB_TX_SET_DAV_LOW;
				if (NRFD_PORT == 1) tx_state = GPIB_TX_SET_DAV_LOW;
				//if (PIR0bits.TMR0IF) rx_state = GPIB_TX_SET_DAV_LOW;

				break;
			case GPIB_TX_SET_DAV_LOW:
                //trigger(1);
				DAV_TRIS = 0;									//		DAV = 0;
				tx_state = GPIB_TX_WAIT_FOR_NDAC;
				count = TIMEOUT;
				//TMR0L = 0;										// Alternative timeout on TMR0 overflow
				//PIR0bits.TMR0IF = 0;
				break;
                
			case GPIB_TX_WAIT_FOR_NDAC:
                NDAC_TRIS = 1;
                //trigger(NDAC_PORT+1);
                //dump();
				if(NDAC_PORT == 1 || count--<=0) tx_state = GPIB_TX_SET_DAV_HIGH;
				if(NDAC_PORT == 1) tx_state = GPIB_TX_SET_DAV_HIGH;
                //tx_state = GPIB_TX_SET_DAV_HIGH;
				//if (PIR0bits.TMR0IF) rx_state = GPIB_TX_SET_DAV_HIGH;
                //xprintf(" %d %d %d -", NDAC_PORT, tx_state, j);
				break;
			case GPIB_TX_SET_DAV_HIGH:
                //trigger(4);
                EOI_TRIS = 1; 										// EOI false
				//
                DAV_TRIS = 1;										// DAV high
				DIO_TRIS=0xff; 										// data lines tristate
				//
				// More data?
				//
				if (j < len) tx_state = GPIB_TX_CHECK;
				else tx_state = GPIB_TX_DONE;
				break;
			case GPIB_TX_ERROR:
//#if SDEBUG >= 1
				xprintf("{ERR}");
//#endif
				tx_state = GPIB_TX_DONE;
				break;
		}
	}
	// All sent							
	DAV_TRIS = 1;													// DAV, NRFD etc threestate (input)
	NRFD_TRIS = 1;
	NDAC_TRIS = 1;
	ATN_TRIS = 1;
	EOI_TRIS = 1;												
	DIO_TRIS = 0xff;												// Data threestate (input)
    //
    //xprintf("TX Exit");
    //trigger(3);
    LATAbits.LATA7 = 0;     // Disk LED
    //PIR0bits.IOCIF = 0; // Clear interrupt on change flag 
    //PIE0bits.IOCIE = 1;     // Enable interrupt-on-change
    IOCBNbits.IOCBN0 = 1;   // enable Interrupt on negative-going edge DAV
                            // to enable reception
    return j;
}
//
// Assert SRQ to get some attention
// state = true -> assert SRQ, state = false -> deasert SRQ
//
void gpib_srq(unsigned char state)
{
#if defined(SDEBUG)
		if(gpib_debug & GPIB_DBG_SPOLL)
		{
			xprintf("%s ","[SRQ on]");
}
#endif
	if(state == true)
	{
		//TRISAbits.TRISA2 = 0;		// Assert SRQ
		//PORTAbits.RA2 = 0;
        SRQ_TRIS = 0;
        SRQ_PORT = 0;
#if defined(SDEBUG)
		if(gpib_debug & GPIB_DBG_SPOLL)
		{
			xprintf("%s ","[SRQ on]");
}
#endif
	}
	else
	{
		//TRISAbits.TRISA2 = 1;		// Deassert SRQ
		//PORTAbits.RA2 = 1;
        SRQ_TRIS = 0;
        SRQ_PORT = 0;
#if defined(SDEBUG)
		if(gpib_debug & GPIB_DBG_SPOLL)
		{
			sio_puts("[SRQ off]");
}
#endif
	}
}
//
// GPIB receive interrupt ISR. Called by DAV or ATN going low. 
// DAV - Stores GPIB bytes and status in circular buffer
// ATN - Wiggles NRFD and NDAC
// 
// Does minimal processing of addressing to handle listening/talking status as
// well as checking for buffer full
//
// Refer to states on figure 2.3 data byte transfer timing on pg 13 of "Tutorial Description of the Hewlett-Packard Interface Bus"
//
void  __interrupt(high_priority) myHighIsr(void)
{
    int i;
    unsigned char byte;
    unsigned char lastbyte;
    unsigned char status;
    // Check IOCBF for ATN - End up here if ATN changes state
    //
    if(IOCBFbits.IOCBF4)
    {
        if(IFC_PORT == 0)
        {
            talking = 0;
            listening = 0;
        }
        if(ATN_PORT == 0 || listening)
        {
            NDAC_TRIS = 0;			// Set NDAC low if command or if we are listening
            //NRFD_TRIS = 1;          // Set NRFD high, this is State -1
            // Check if buffer full
            //
            if(RxFifo.ct>=BUFFER_SIZE)
            {
                NRFD_TRIS = 0;			// Set NRFD low to indicate that we  
            }                           // cannot accept more data
            else 
                NRFD_TRIS = 1;			// Set NRFD high, can accept data
        }
        if(ATN_PORT == 1 && !listening) NDAC_TRIS = 1;         // Not command and not listening, set NDAC high
        IOCBFbits.IOCBF4 = 0;
    }
    // Check IOCCF for DAV
    // State 0 - DAV went low
    if(IOCBFbits.IOCBF0)
    {
        //
        // State 1 - Pull NRFD down
        //
        if(ATN_PORT == 0 || listening) NRFD_TRIS = 0;			// Set NRFD low if command or if we are listening
        //
        // Accept databyte
        //
        lastbyte = byte;
        byte = ~DIO_PORT;
        status = ~STATUS_PORT;
        i = RxFifo.ct;				/* Number of bytes in the FIFO */
        if (i < BUFFER_SIZE) {		/* Skip if FIFO is full */
            RxFifo.ct = ++i;
            i = RxFifo.wi;
            RxFifo.buff[i] = byte;          // Store received data into the FIFO
            RxFifo.cbuff[i++] = status;     // Store the control byte
            RxFifo.wi = i % BUFFER_SIZE;	/* Next write ptr */
        }
        //
        // State 2 - Set NDAC high
        //
        if(ATN_PORT == 0 || listening) NDAC_TRIS = 1;			// Set NDAC high
        NDAC_TRIS = 1;
        // Wait here for DAV to go high
        while(DAV_PORT == 0)
            ;                                                               // Todo: fix timeout
        //
        // State 3 - DAV went high 
        //
        if(ATN_PORT == 0 || listening)
        {
            NDAC_TRIS = 0;			// State 4 - Set NDAC low
            //NRFD_TRIS = 1;			// State 5 - Set NRFD high
        }
        else
        {
            NDAC_TRIS = 1;			// Set NDAC high
            //NRFD_TRIS = 1;			// Set NRFD high        
        }
        //
        // Check if buffer full
        //
        if(RxFifo.ct>=BUFFER_SIZE)
        {
            NRFD_TRIS = 0;			// Set NRFD low to indicate that we  
        }                           // cannot accept more data
        else 
            NRFD_TRIS = 1;			// Set NRFD high, can accept data
        //
        // Process addressing commands if ATN was asserted with the databyte
        //
        //if(Disk.sector == 4 && Disk.cyl == 32) trigger(1);
        //if (ATN_PORT == 0) // 
        if (status & ATN_BIT)
        {
            //if(Disk.sector == 4 && Disk.cyl == 32) trigger(2);
            switch(byte)
            {
                case MLA: 
                    listening = true;
                    talking = false;
                    break;
                case MTA: 
                    talking = true;
                    listening = false;
                    break;
                case UNL:
                    listening = false;
                    break;
                case UNT:
                    //if(Disk.sector == 4 && Disk.cyl == 32) trigger(3);
                    talking = false;
                    break;
                case SDC:                   // Selected device clear
                case DCL:                   // Device clear
                    if(listening)
                    {
                        talking = 0;
                        listening = 0;
                    }
            }
        }
        IOCBFbits.IOCBF0 = 0;
    }
    PIR0bits.IOCIF = 0; // Clear interrupt on change flag 
}

int gpib_test (void)
{
	return RxFifo.ct;	/* Returns number of bytes in the Rx FIFO */
}


/* Get a byte plus control byte from GPIB */

BYTE gpib_getc (char *cbyte)
{
	BYTE d,c;
	int i;

	while (!RxFifo.ct) ;		/* Wait while Rx FIFO empty */

	i = RxFifo.ri;				/* Get a byte from Rx FIFO */
	d = RxFifo.buff[i];
    c = RxFifo.cbuff[i++];
	RxFifo.ri = i % BUFFER_SIZE;
	di();
	RxFifo.ct--;
	ei();
    *cbyte = c;
    //
    // The ISR might have set NRFD low to indicate buffer full
    // check if space available and change accordingly if listening
    // also if ATN asserted
    // TODOD: Check for 90% full instead!
    //
    if(RxFifo.ct<BUFFER_SIZE && ((ATN_PORT == 0 || listening)))
    //if(RxFifo.ct<BUFFER_SIZE && ( listening))
        NRFD_TRIS = 1;			// Set NRFD high, can accept data
	return d;
}
void dump(void)
{
    xprintf("\r\n %2X %2X %2s %2s %2s %2s %2s %2s %2s %2s %4s %4s %4s %4s %4s %4s %4s %4s ",
            (~PORTD&0xff), PORTB,
            PORTDbits.RD0 == 0 ? "D0":"  ",
            PORTDbits.RD1 == 0 ? "D1":"  ",
            PORTDbits.RD2 == 0 ? "D2":"  ",
            PORTDbits.RD3 == 0 ? "D3":"  ",
            PORTDbits.RD4 == 0 ? "D4":"  ",
            PORTDbits.RD5 == 0 ? "D5":"  ",
            PORTDbits.RD6 == 0 ? "D6":"  ",
            PORTDbits.RD7 == 0 ? "D7":"  ",
            DAV_PORT == 0 ? "DAV ":"  ",
            NRFD_PORT == 0 ? "NRFD":"  ",
            NDAC_PORT == 0 ? "NDAC":"  ",
            IFC_PORT == 0 ? "IFC ":"  ",
            SRQ_PORT == 0 ? "SRQ ":"  ",
            ATN_PORT == 0 ? "ATN ":"  ",
            EOI_PORT == 0 ? "EOI ":"  ",
            REN_PORT == 0 ? "REN ":"  ");     
}

//
// Dump an data-area in human-readable form
//
void DumpData(char *Ptr,int length, char *label){
int i,j;
char ch;
	xprintf("\r\n%s %X %X\r\n",label, Ptr, length);
	for(j=0;j<80&&(j*16<length);j++){
		xprintf("\r\n");
		for(i=0;i<16 && (i+j*16<=length);i++)
			{
				ch = *(Ptr+i+j*16);
				xprintf(" %02X",ch&0xFF);
			}
		xprintf(" | ");
		for(i=0;i<16 && (i+j*16<=length);i++)
			{
				if(*(Ptr+i+j*16)>' ') xprintf("%c",(*(Ptr+i+j*16))&0xFF); 
				else xprintf(".");
			}
	}
	xprintf("\r\n");
}
int gpib_flush()
{
    RxFifo.ri=0;
    RxFifo.wi=0;
    RxFifo.ct=0;
    return 0;
}

//
// This is just to help us debug timer-critical parts. Wiggle a6 n times
//
void trigger(char n)
{
    while(n--)
    {
        LATAbits.LATA6 = 1;
        LATAbits.LATA6 = 0;    
    }
    NOP();
    NOP();

}