//-------------------------------------------------------------------------
//  main.c - HPDisk main function
//
// (c) 2010 - 2020 Anders Gustafsson <anders.gustafsson@pedago.fi> 
//
// Disk emulator for SS/80 disks. New improved version of HPDisk
// Reception is now interrupt-driven
//-------------------------------------------------------------------------

// 18F46K40 
// This chip needs a compiler workaround:
// --ERRATA=NVMREG
//  
// 
// The proper way to do it with XC8 version 1.40 and later is the following:
// Enter +nvmreg in the Project Properties XC8 global options->XC8 Linker->Additional Options->Errata box
// That is, it's a Linker option, not a Compiler option.
//
// From: http://www.microchip.com/forums/m965517.aspx
//

// History
// 20190526 0.1 First working version
// 20190528 0.2 Now reads and writes correctly it seems
// 20190530 0.3 Also does Amigo, albeit only when SDEBUG is 1 or above
// 20191202 0.4 Works to some degree with HP 9816. Still lacks support for unbuffered read/writes
// 20191206 0.5 Works, but not yet writes. Can define filenames for the 8 LIF files
// 20191207 0.6 BOOTS BASIC 2.0 and 4.0. Can store and load programs from BASIC 2.0
// 20200104 0.7 Allows for 16 filenames, PORTE 1,2,3 plus RA5
// 20200112 0.8 Support for 4 physical disks, DISK0..DISK4, disk 0 is selectable
//              from 16 image files. Disk 1..3 are selectable through the cfg file
//              this number of drives is defined in defines.h
// 20200123 0.9 Added code to read time from a DS2417
// 20200129     Code to read/write DS2417 finished.
// 20200131     Corrected an error where the GPIB handshake could get stuck receiving
// 20200201     Activated WDT
// 20200219 0.10 Support for display/key/rtc daughterboard, DSKSEL 3
//               Stores last selected disk number i EEPROM
//
// xc8 v2.10
//
// Onewire: https://www.maximintegrated.com/en/design/technical-documents/app-notes/1/126.html
//
#define VERSION "0.10"

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.
//; CONFIG1L
#pragma config  FEXTOSC = OFF               //; External Oscillator mode Selection bits (Oscillator not enabled)
#pragma config  RSTOSC = HFINTOSC_64MHZ     //; Power-up default value for COSC bits (HFINTOSC with HFFRQ = 64 MHz and CDIV = 1:1)
#pragma config  CLKOUTEN = OFF
#pragma config  WDTE = OFF
#pragma config  WDTCPS = WDTCPS_14          // Timeout in 16 sec

#include <xc.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include "main.h"
#include "sio.h"
#include "xprintf.h"
#include "gpib.h"
#include "defines.h"
#include "Globals.h"
#include "ff.h"
#include "ss80.h"
#include "amigo.h"
#include "ds2417.h"
#include "display.h"
#include "eeprom.h"
//#include "integer.h"
//
// States for startup
//
#define START 10
#define WAIT_FOR_CARD 20
#define MOUNT 30
#define READ_CONFIG 35
#define OPEN_EXISTING 40
#define FILE_SEEK 45
#define OPEN_NEW 50
#define FILE_OPEN 60
#define WAIT_FOR_DATA 70
#define PROCESS_DATA 80
#define WRITE_DATA 90
#define CLOSE_FILE 100
#define WAIT_FOR_EJECT 110
#define NEW_HOUR 120
#define POWER_FAIL 130
#define NEW_DAY 140

unsigned char idbyte1;                      // Bytes returned by IDENT GPIB commands
unsigned char idbyte2;

char filenames[16+MAXDISK+1][MAXFN] =       // Name of LIF image files 0..
{
    "/lifdata0.bin ",
    "/lifdata1.bin ",
    "/lifdata2.bin ",
    "/lifdata3.bin ",
    "/lifdata4.bin ",
    "/lifdata5.bin ",
    "/lifdata6.bin ",
    "/lifdata7.bin ",
    "/lifdata8.bin ",
    "/lifdata9.bin ",
    "/lifdataa.bin ",
    "/lifdatab.bin ",
    "/lifdatac.bin ",
    "/lifdatad.bin ",
    "/lifdatae.bin ",
    "/lifdataf.bin ",                       // Last file for disk 0
    "",                                     // Disk 1 etc
    "",
    ""
};
char cfgfile[] = "hpdisk.cfg";              // Name of config file

FIL File[MAXDISK+1];

char *buffer = gpib_tx_buff;                // Disk buffer used by ReadConfig())

volatile BYTE state = 0;					// State machine state needs to be global

volatile BYTE gpib_debug = 255 -GPIB_DBG_RX - GPIB_DBG_TX;				// Debug level for debug builds, see defines!

unsigned char gpib_flags;                   // Prologix gpib flags = device, EOI=1
unsigned char dsksel = 0;                   // How we select disk
unsigned char diskswitch = 0;               // Value read from disk switch
unsigned char rtc = 0;                      // True if we have a DS2417 RTC
unsigned char temp_diskno = 0;              // This is the temporary disk select
//
FATFS fs;                       // File system object
FRESULT rc;
unsigned char MMC; 				// Memory card status
DWORD fpos;						// File position

BYTE ds;
DWORD MagicVector = 0;          // If nonzero indicates what sector should
                                // be redirected to RTC

//
int Mount_disk(void);
int Read_config(void);
void Error_blink(int code);
static void put_rc (FRESULT rc);
unsigned char ReadSwitch(void);
unsigned char DiskSelect(void);

BYTE disk_status (
  BYTE pdrv     /* [IN] Physical drive number */
);
//
// Loop forever checking input and output
//
void main(void)
{
	unsigned char pending = 0;
	int rc, rc2;
    unsigned int i;
    //
    char b, b2,c;
    unsigned char key;
    int status;
    UINT nbytes; // for testing only
    diskno = DATAEE_ReadByte(0);
    diskswitch = 0;

    DS2417 t;
	//
	TRISE = 0x07;				// Port E input for disk image file select
	//
	protocol = 1;				// Default protocol: SS/80
	debuglevel = 0;				// Default loglevel
	//
	// Set ports initially
	//
	ADCON1=0x0f;				// Turn off all analogue
    ANSELA = 0;
    ANSELB = 0;
    ANSELC = 0;
    ANSELD = 0;
    ANSELE = 0;
 
	TRISA=0xff;                  // Ports initially inputs
	TRISB=0xff;
	TRISC=0xff;                  // Listener NRFD, NDAC active  0;

	TRISCbits.TRISC7 = 1;

    TRISAbits.TRISA0 = 0;               // PP Enable
	TRISAbits.TRISA1 = 0;				// "Listening" LED
	LATAbits.LATA1 = 1;					// Led on
	gpib_flags = FLAGS_EOI;
	sio_setup();
	//
	TRISAbits.TRISA2 = 0;				// "Talk" LED
	LATAbits.LATA2 = 1;					// Led on	
	//
    // Trigger output pins for debugging
    //
    TRISAbits.TRISA6 = 0;				// T1  output
    LATAbits.LATA6 = 0;	
    //LATAbits.LATA6 = 1;	
    //__delay_us(5);
    //LATAbits.LATA6 = 0;	
    TRISAbits.TRISA7 = 0;				// T2  output, activity LED
    LATAbits.LATA7 = 1;                 // Light LED, will be on until ready
    //
    DIO_TRIS = 0xff;
    STATUS_TRIS = 0xff;
    xprintf("\r\nHPDisk Mk II (18F46K40) %s %s %s Starting...\r\n", VERSION, __DATE__, __TIME__);
    //
    Globals_IO_init();  	// Set in and outputs for SD-Card
    //
    while(CD_PORT)
    {
        xprintf("\r\nNo card %d", PORTC);
        __delay_ms(500);
        CLRWDT();
    }
    diskno = ReadSwitch();
    xprintf("\r\nDisk select number %d", diskno);
    //
    ClrWdt();
    __delay_ms(500);
    //
    // Set up timer 1 to be used for timing of debug messages
    // 64Mhz/4/8 = 0.5ns per tick, 32.8ms max if using FOSC/4
    //
    TMR1CLKbits.CS = 1;     // 64Mhz/4/8 = 0.5ns per tick
    
    T1GCONbits.GE = 0;
    PIE4bits.TMR1IE = 0;    // No interrupts
    T1CONbits.CKPS = 0;     // 3 = 1:8 prescale
    T1CONbits.RD16 = 1;     // 16-bit operation
    T1CONbits.ON = 1;       // Timer on
    //
    // Init interrupt on change for PORTB.0 (DAV) PORTC.4 (ATN))
    PIE0bits.IOCIE = 1;     // To allow individual port pins to generate an interrupt, the IOCIE bit of the PIE0 register must be set.
    IOCBNbits.IOCBN0 = 1;   // Interrupt on negative-going edge DAV
    IOCBNbits.IOCBN4 = 1;   // ATN negative going
    IOCBPbits.IOCBP4 = 1;   // ATN positive going
    INTCONbits.GIE = 1;     // Enable interrupts
    IPR0bits.IOCIP = 1;     // Make interrupt on change high priority
	gpib_init();
    
    listening = 0;
    talking = 0;
    pending = 0;
//    while(1)
//    {
//        EnablePP();
//        DisablePP();
//        trigger(1);
//    }
    lcd_init();
    lcd_init();
    writedata(0x34,0);      //full height
    lcd_puts("HPDisk ");
    lcd_puts(VERSION);
    writedata(0x39,0);      //normal
    //
    DisablePP();
    while(1)
    {
		ClrWdt();
		//
        
        rc=Mount_disk();
        SS80_init();
        amigo_init();
        //
        writedata(0x80,0);
        lcd_puts(filenames[diskno]);
        //
        // Jan 1 1970 is Julian date 2440587.5
        // or times 86400 = 2108667600000
        // 0x427EAF6621C80000 in IEEE754
        // http://weitz.de/ieee/
        // https://github.com/MalcolmMcLean/ieee754/blob/master/ieee754.c
        // Check if user has indicated we have an RTC connected
        if(rtc)
        {
            //
            xprintf("\r\nReading RTC\r\n");
            t = (DS2417) DS2417_read();
            if(t.time==0)
            {
                xprintf("\r\nGot %lu  - resetting", t);
                DS2417_init(10);
            }
            for(int i = 0; i<4;i++)
            {
                xprintf("%2X", t.ch[i]);
            }
            xprintf("\r\n");
            //
            xprintf("Looking for an RTC file\r\n");
            MagicVector = FindFile(File[0],"RTC");
            if(MagicVector > 0)
            {
                xprintf("Found an RTC file at %X\r\n", MagicVector);
                rc=f_lseek(&File[0],(MagicVector+1)*256);
                rc=f_read(&File[0],buffer,8,&nbytes );
                for(int i = 0; i<8;i++)
                {
                    xprintf(" %X", buffer[i]);
                }
                //test = (long int) *buffer;
                //xprintf("\r\nValue: %X\r\n", test);
            }
        }
        //
        //
        if(debuglevel & 128)        // Just loop and display port status
        {
            TRISC = 0xff;           // All input for testing
            while(1)
            {
                LATAbits.LATA1 = 0;	
                LATAbits.LATA1 = 1;	
                __delay_ms(500);
                dump();
            }
        }
        
        if(protocol)            // SS80
        {
#if defined(HP9122D)
            xprintf("\r\nSS/80 9122D\r\n");
#endif
#if defined(HP9134L)
            xprintf("\r\nSS/80 9134L\r\n");
#endif
            idbyte1 = SS80ID1;
            idbyte2 = SS80ID2;
        }
        else					// Amigo
        {
							// Amigo drives
#if defined(HP9121D)
            xprintf("\r\nAmigo 9121D\r\n");
#endif
            idbyte1 = AMIGOID1;
            idbyte2 = AMIGOID2;
        }
        
        rc2 = 0;
        b = 0;
        //
        //
        // Open the three other diskfiles if defined
        //
        for(int i = 1; i<=MAXDISK;i++)
        {
            if(strlen(filenames[i+16]) > 2)
            {
                xprintf("\r\nOpening %s  ", filenames[i+16]);
                rc = f_open(&File[i], filenames[i+16], FA_OPEN_EXISTING + FA_WRITE+ FA_READ); //FA_CREATE_NEW + FA_WRITE);
                put_rc(rc);
            }
        }
        //
        // dsksel designates newer type of select. Internal pullups and switch to ground
        //
        if(dsksel == 2)
        {
            WPUAbits.WPUA5 = 1;
            WPUE = 0x07;
        }
        LATAbits.LATA7 = 0;
        writedata(0x80+0x40,0);
        lcd_clrln();
        writedata(0x80+0x40,0);
        lcd_puts("Ready");
        
        while(CD_PORT == 0 && rc2 != 5) // SDCard CD is low when card present, so loop here and fall through if card removed
		{
            //dump();
            //
            // We might be started or reset whilst the host is keeping ATN
            // asserted. This will not be caught by the ISR so catch it here.
            //
            if(ATN_PORT == 0)
            {
                NDAC_TRIS = 0;			// Set NDAC low if command so that the
                                        // host sees us.
            }
            if(gpib_test())     // We have a character from GPIB
            {
                b2 = b;
                b = gpib_getc(&c);
                xprintf("[%2X %d]", b, PP_PORT);    // Seems to need this for a delay
                //
                if((c & ATN_BIT) && b == 0x60 && b2 == 0x5f)	// GPIB Ident sequence?
                {
                    
                #if SDEBUG >= 1
                    xprintf("<ident> %X %X %d %d", b, b2, NRFD_PORT, NDAC_PORT);
                #endif
                    //trigger(2);
                    gpib_flags = gpib_flags | FLAGS_EOI;
                    gpib_tx_buff[0] = idbyte1;			// Identify ourselves as a HP 9121 drive
                    gpib_tx_buff[1] = idbyte2;			// $0104 or 9122 $0222
                    talking = 1;
                    sendgpib(2);
                    talking = 0;
                    //listening = 0;
                    gpib_flags = gpib_flags & ~FLAGS_EOI;
                    gpib_ptr = 0;
                    gpib_flush();
                    b = 0;
                    continue;
                }
                //xprintf("%X ", b );
#if SDEBUG > 9
                xprintf("\r\n %02X %c %4s %4s %4s %4s %4s ",
                b,b,
                c&IFC_BIT ? "IFC ":"  ",
                c&SRQ_BIT ? "SRQ ":"  ",
                c&ATN_BIT ? "ATN ":"  ",
                c&EOI_BIT ? "EOI ":"  ",
                c&REN_BIT ? "REN ":"  ");
#endif
                //
                if(protocol)				// SS/80
                {
                    status = SS80_char(b,c&ATN_BIT,0);  // Todo: IFCFLAG
                    if(status == SS80_CMD_NOP) Nop(); 
                    if(status == SS80_CMD_CLEAR) SS80_cmd_clear();
                    if(status == SS80_CMD_DESCRIBE) SS80_cmd_describe();
                    if(status == SS80_CMD_SEND_STATUS) SS80_cmd_send_status();
                    if(status == SS80_CMD_BUFFERED_READ) rc2 = SS80_cmd_buffered_read();
                    if(status == SS80_CMD_BUFFERED_WRITE) rc2 = SS80_cmd_buffered_write();
                    if(status == SS80_CMD_FORMAT)  SS80_cmd_format();
                }
                else						// Amigo
                {
                    status = amigo_char(b,c&ATN_BIT,0);
                    if(status == AMIGO_CMD_CLEAR) amigo_cmd_clear();
                    if(status == AMIGO_CMD_SEND_STATUS) amigo_cmd_send_status();
                    if(status == AMIGO_CMD_SEEK) amigo_cmd_seek();
                    if(status == AMIGO_CMD_BUFFERED_READ) amigo_cmd_buffered_read();
                    if(status == AMIGO_CMD_DSJ) amigo_cmd_dsj();
                    if(status == AMIGO_CMD_BUFFERED_WRITE) amigo_cmd_buffered_write();
                    if(status == AMIGO_CMD_WAKEUP) amigo_cmd_wakeup();
                    if(status == AMIGO_CMD_UNBUFFERED_READ) amigo_cmd_unbuffered_read();
                    if(status == AMIGO_CMD_UNBUFFERED_WRITE) amigo_cmd_unbuffered_write();
                    if(status == AMIGO_CMD_FORMAT) amigo_cmd_format();
                }
            }
            else    // Nothing to do, check if we want to change disk
                    // Note that this needs to be done rather quickly or it will
                    // upset IDENT
            {
                diskswitch = DiskSelect();
                 //
                // Has it changed?
                if(dsksel >= 1 && diskswitch!= diskno)
                {
                    LATAbits.LATA7 = 1;     // Light LED
                    writedata(0x80+0x40,0);
                    lcd_clrln();
                    writedata(0x80+0x40,0);
                    lcd_puts("Loading...");
                    diskno = diskswitch;

#if SDEBUG >= 1
                    xprintf("Disk change: %d  \r\n", diskno);
#endif
                    break;
                }
            }
        } // while CD_PORT
        while(CD_PORT)
        {
            xprintf("\r\nNo card %d", PORTC);
            __delay_ms(500);
            ClrWdt();
            writedata(0x80+0x40,0);
            lcd_clrln();
            writedata(0x80+0x40,0);
            lcd_puts("No card");
        }
   __delay_ms(500);
   talking = 0;
   listening = 0;
   f_mount(NULL,"",0);
   }
} // While 1
//
// Read config from HPDISK.CFG on SD-card
//
// A straight textfile with data in the form of lines
// containing keys and values, ie:
//
// PROTO 0
// DEBUG 1
//
// Note! We use gpib_tx_buff for temporary storage
int Read_config(void)
{
char key[PARMLEN];
char val[PARMLEN];
char filekey[MAXFN];
UINT nbytes;
int c,n, i,j,k;
    //
    // OK, we seem to have a config file. Parse it to get values
    //
    rc=f_read(&File[0],buffer,512,&nbytes );
	xprintf("\r\nRead %d bytes", nbytes);
	//
	n=0;
	c=0;
	for(j=0;j<PARMLEN;j++)
	{
		key[j]=0;
		val[j]=0;
	}
	for(i=0;i<nbytes;i++)
	{
		if(buffer[i] == 0x0d)
		{
#if SDEBUG > 1
		xprintf("\r\nK:%s V:%s", key, val);
		//
#endif
		if(!strcmp("PROTO", key))
			protocol = atoi(val) & 0xff;
		if(!strcmp("DEBUG", key))
			debuglevel = atoi(val) & 0xff;
		if(!strcmp("DSKSEL", key))
			dsksel = atoi(val) & 0xff;
        if(!strcmp("RTC", key))
			rtc = atoi(val) & 0xff;
        //
        // Disk 0
        // Disk image names as NAMEx, where x is 0..15
        //
        for(k=0;k<16;k++)
        {
            sprintf(filekey,"NAME%X",k);
            if(!strcmp(filekey, key))
                strncpy(filenames[k],val, MAXFN);
        }
        // Disk 1..3
        //
        for(k=1;k<=MAXDISK;k++)
        {
             sprintf(filekey,"DISK%X",k);
            if(!strcmp(filekey, key))
                strncpy(filenames[k+16],val, MAXFN);           
        }
//		if(!strcmp("QSTAT", key))
//			qstat = atoi(val) & 0xff;
		//	
		for(j=0;j<PARMLEN;j++)
		{
			key[j]=0;
			val[j]=0;
		}
			c=0;
			n=0;
		}
		else if(buffer[i] == 0x0a)
		{
		}
		else if(buffer[i]==' ')
		{
			c=1;
			n=0;
		}
		else
		{
			if(c) val[n] = buffer[i];
			else key[n] = buffer[i];
			if(n < PARMLEN) n++;
		}
	}
    xprintf("\r\nDsksel is %d ", dsksel);
    xprintf("\r\nProto is %d ", protocol);
    xprintf("\r\nDebug is %d ", debuglevel);
    for(k=0;k<16;k++)
    {
        xprintf("\r\nFile %X is %s ", k, filenames[k]);
    }
    for(i=1;i<=MAXDISK;i++)
    {
        xprintf("\r\nDisk %X is %s ", i, filenames[16+i]);
    }
    xprintf("\r\n");
	return 0;
}
//
// Blink the error code. Blink listening LED <code> short, followed by one long
//
void Error_blink(int code)
{
	listening = 0; 
	__delay_ms(1000);
	listening = 1; 
	__delay_ms(1000);
	while(code--)
	{
		listening = 0; 
		__delay_ms(300);
		listening = 1;
		__delay_ms(300);
	}
	listening = 0; 
	__delay_ms(2000);
}

static void put_rc (FRESULT rc)
{
	const char *p;
	FRESULT i;
	static const char str[] =
		"OK\0DISK_ERR\0NOT_READY\0NO_FILE\0NOT_OPENED\0NOT_ENABLED\0NO_FILE_SYSTEM\0";

	for (p = str, i = 0; i != rc && *p; i++) {
		while(*p++) ;
	}
	xprintf("rc=%u FR_%s\r\n", rc, p);
}

int Mount_disk(void)
{
int state = START;
int retries = 5;

while (retries > 0)
    {
    switch(state)
    {
        case START:
            state = WAIT_FOR_CARD;
        break;
        case WAIT_FOR_CARD:                     // Wait for user to insert card
            sio_puts("\r\nWaiting for card ");
            if(CD_PORT==0) state = MOUNT;
            __delay_ms(500);
            //listening = !listening;             // Wiggle LED
        break;
        case MOUNT:                             // Register work area to the default drive
            //LED1_LAT = !LED1_LAT;             // Wiggle LED
            sio_puts("\r\nMounting ");               // Try to mount
            rc = f_mount(&fs, "", 0);
            put_rc(rc);
            if(rc == FR_OK) state = READ_CONFIG;
            __delay_ms(500);
        break;
        case READ_CONFIG:
            //LED1_LAT = !LED1_LAT;             // Wiggle LED
            sio_puts("\r\nRead config ");
            rc = f_open(&File[0], "hpdisk.cfg", FA_OPEN_EXISTING + FA_READ);
            put_rc(rc);
            if(rc == FR_OK)
            {
                Read_config();
                rc = f_close(&File[0]);
            }
            put_rc(rc);
            state = OPEN_EXISTING;
            //
            // Change initial filename in case disk selection via porte is active
            //
            if(dsksel==0)
            {
                diskno = 0;
                sprintf(filenames[0],"/lifdata.bin");
            }
            else
                xprintf("\r\nDsksel is %d, number is %d ", dsksel, diskno);
            xprintf("\r\nFilename %s ", filenames[diskno]);
            //
            __delay_ms(500);
        break;
        case OPEN_EXISTING:                     // Open an existing file
            //LED1_LAT = !LED1_LAT;             // Wiggle LED
            xprintf("\r\nOpening existing %s ", filenames[diskno]);
            //
            rc = f_open(&File[0], filenames[diskno], FA_OPEN_EXISTING + FA_WRITE+ FA_READ); //FA_CREATE_NEW + FA_WRITE);
            //put_rc(rc);
            if(rc == FR_OK) state = FILE_OPEN;
            if(rc == FR_NO_FILE) state = OPEN_NEW;
            __delay_ms(500);
        break;
        case FILE_SEEK:
        break;
        case OPEN_NEW:                          // No existng,  Open a new image file
            //LED1_LAT = !LED1_LAT;             // Wiggle LED
            xprintf("\r\nOpening new %s ", filenames[diskno]);

            rc = f_open(&File[0], filenames[diskno], FA_CREATE_NEW + FA_WRITE + FA_READ);
            //put_rc(rc);
            if(rc == FR_OK) state = FILE_OPEN;
            __delay_ms(500);
        break;
        case FILE_OPEN:
            //LED1_LAT = !LED1_LAT;             // Wiggle LED
            xprintf("\r\nFile %s is open ", filenames[diskno]);
            state = WAIT_FOR_DATA;
            //break;
            return 1;
        } // switch
    if(state != WAIT_FOR_CARD) --retries;
    } // while
return 0;
}
unsigned char DiskSelect(void)
{
    char d = diskno;
    char key;
    switch(dsksel)
    {
        case 1:
        case 2:
            d = ReadSwitch();
            if(d != diskno)
                __delay_ms(500);
            break;
        case 3:
            key = readkeys();

            if(key>0)
            {
                //xprintf("Key %X %d  ",key, d);
                if(key & 0x01) ++temp_diskno;       // Up/down keys
                if(key & 0x02) --temp_diskno;
                temp_diskno = temp_diskno & 0x0f;
                //printf(string,"Key %X %d",key, d);
                //writecom(0x80);
                //writedata(d+'a');
                writedata(0x80+0x40,0);
                lcd_clrln();
                writedata(0x80+0x40,0);
                lcd_puts(filenames[temp_diskno]);
                __delay_ms(300);
                if(key & 0x04)  // Enter
                {
                    DATAEE_WriteByte(0, temp_diskno);
                    d = temp_diskno;
                    writedata(0x80,0);
                    lcd_clrln();
                    writedata(0x80,0);
                    lcd_puts(filenames[temp_diskno]);
                    writedata(0x80+0x40,0);
                    lcd_clrln();
                }
            }
            break;
    }
    return d;
}
//
// read the disk selector switch, taking into account that it can be
// wired as pulldowns (DSKSEL = 1) or weak pullups (DSKSEL = 2)
unsigned char ReadSwitch(void)
{
    unsigned char sw;
    sw = (PORTE & 0x07);
    TRISAbits.TRISA5 = 1;
    if(PORTAbits.RA5)
        sw = sw+8;
    //
    if(dsksel == 2) sw = ~sw & 0x0f;
    return sw;
}