
#include <stdio.h>
#include <string.h>	// for memset
#include "defines.h"
#include "Globals.h"
#include "gpib.h"
#include "db_utils.h"
#include "sio.h"
#include "xprintf.h"    // Less memory footprint than printf())
#include "lifdisk.h"
#include "ss80.h"
#include "ff.h"
#include "ds2417.h"

//
// SS80 disk handling. See "Subset 80 from Fixed and flexible disc drives" 5958-4129 
//
// HP 9122 SS80 disk is stored on the SD card as a disk image cylinder - head - sector
//
// Docs:
// LIF filesystem http://www.hp9845.net/9845/projects/hpdir/#lif_filesystem

// ----------------------------------------- Structures used for testing ------------------------------------------
// HP/LIF data is BIG endian
//
// Volume label structure
//
#if defined(HP9122D)
ControllerDescriptionType ControllerDescription = {
0b10000000,
0b00000001,	// One unit
0x02,		// Transfer rate 744 = 0x2e8
0xe8,
0x05		// Controller type
};

UnitDescriptionType UnitDescription = {
0x01,		// Unit type = floppy
0x09,		// Type 9122
0x12,
0x20,
0x01,		// Bytes per block = 256
0x00,
0x01,		// Number of buffered blocks
0x00,		// Burst size = 0
0x17,		// Block time in us = 5888 = 0x1700
0x00,
0x00,		// Continous transfer rate = 45
0x2d,
0x11,		// Optimal retry = 4500 = 0x1194
0x94,
0x20,		// Access time = 8400 = 0x20d0
0xd0,
0x0f,		// Maximum interleave
0x00,		// Fixed volume byte (no fixed volumes)
0x01		// Removable volume byte = 1
};

VolumeDescriptionType VolumeDescription =
{
0x00,		// Max cylinder
0x00,
0x00,
0x00,		// Max head
0x00,		// Max sector
0x00,
0x00,		// Max single vector address = 2463 = 99f
0x00,
0x00,
0x00,
0x09,
0x9f,
0x02		// Current interleave = 2
};
#elif defined(HP9134L)
ControllerDescriptionType ControllerDescription = {
0b10000000,
0b00000001,	// One unit
0x02,		// Transfer rate 744 = 0x2e8
0xe8,
0x05		// Controller type
                                                   };

UnitDescriptionType UnitDescription = {
0x00,		// Unit type = winchester
0x09,		// Type 9134
0x13,
0x40,
0x01,		// Bytes per block = 256
0x00,
0x01,		// Number of buffered blocks
0x00,		// Burst size = 0
0x01,		// Block time in us = 502 = 0x1F6
0xF6,
0x00,		// Continous transfer rate = 140 kb/s
0x8c,
0x11,		// Optimal retry = 4500 = 0x1194 = 45000 = 4,5s worstcase
0x94,
0x11,		// Access time = 4500 = 0x1194
0x94,
0x1f,		// Maximum interleave
0x01,		// Fixed volume byte (1 fixed volumes)
0x00		// Removable volume byte = 0 (fixed)
                                       };

VolumeDescriptionType VolumeDescription =
{
0x00,		// Max cylinder
0x00,
0x00,
0x00,		// Max head
0x00,		// Max sector
0x00,
0x00,		// Max single vector address = 58176 = e340
0x00,
0x00,
0x00,
0xe3,
0x40,
0x07		// Current interleave = 2
};
//#else
//#error You must define a floppy in defines.h
#elif defined(HP7911)
ControllerDescriptionType ControllerDescription = {
0b10000000,
0b00000001,	// One unit
0x02,		// Transfer rate 744 = 0x2e8
0xe8,
0x05		// Controller type
                                                   };

UnitDescriptionType UnitDescription = {
0x00,		// Unit type = winchester
0x09,		// Type 9134
0x13,
0x40,
0x01,		// Bytes per block = 256
0x00,
0x01,		// Number of buffered blocks
0x00,		// Burst size = 0
0x01,		// Block time in us = 502 = 0x1F6
0xF6,
0x00,		// Continous transfer rate = 140 kb/s
0x8c,
0x11,		// Optimal retry = 4500 = 0x1194 = 45000 = 4,5s worstcase
0x94,
0x11,		// Access time = 4500 = 0x1194
0x94,
0x1f,		// Maximum interleave
0x01,		// Fixed volume byte (1 fixed volumes)
0x00		// Removable volume byte = 0 (fixed)
                                       };

VolumeDescriptionType VolumeDescription =
{
0x00,		// Max cylinder
0x00,
0x00,
0x00,		// Max head
0x00,		// Max sector
0x00,
0x00,		// Max single vector address = 58176 = e340
0x00,
0x00,
0x00,
0xe3,
0x40,
0x07		// Current interleave = 2
};
#else
#error You must define a floppy in defines.h
#endif
//
// Amigo: Structure to hold head/cylinder/sector
//
DiskType Disk80 = {0,0,0,0};// Amigo disk addressing
//
// SS/80
// indexed struct to handle multiple disks
//
	FIL* fp;
	int idx, nchr;
	BYTE buf[64];

BYTE Unit;                  // Disk number, index into structure below
BYTE StatusMask[8];         // Mask for masking error bits, set by cmd 3E

struct {
BYTE Volume;
BYTE Address[6];			// SS80 Address, Address[5] = LSB. In blocks
DWORD Vector;				// Vector, ie file position, in blocks
DWORD Maxvector;			// Maximum file position, used for bounds checking. Initialised by SS80_init.
LengthType Length;			// Length, in blocks
BYTE Errors;				// Error byte
} SS80Disk[MAXDISK+1];

//--------------------------------------------------------------------------------------------------------------------
// These are the talking and listening flags for the parser. The bytes are interpreted
// asynchronously from reception and the global talking/listening flags as indicated
// by the leds can very well be something entirely different from when the byte was
// received.
unsigned  char SS80_talking;
unsigned  char SS80_listening;
unsigned  char spoll;
unsigned volatile  char SS80_tstate;    // Transaction state (idle/command/execution/report)
unsigned volatile  int SS80_cstate;		// Command state state
unsigned  char qstat;					// QSTAT byte, 0=OK, 1=Error, 2=Powerup
//
//
// Init SS80 state machine
//
void SS80_init()
{
	Unit = 0;
    for(Unit = 0; Unit < MAXDISK+1; Unit++)
    {
        SS80Disk[Unit].Errors = 0;
        SS80Disk[Unit].Volume = 0;
        SS80Disk[Unit].Vector = 0;
        SS80Disk[Unit].Maxvector = VolumeDescription.V9 * 0x100000000L + 
				VolumeDescription.V10 * 0x1000000L + 
				VolumeDescription.V11 * 0x10000L + 
				VolumeDescription.V12 * 0x100L;
        xprintf("\r\nMax Vector disk %d %ld (%lX) ", Unit, SS80Disk[Unit].Maxvector, SS80Disk[Unit].Maxvector);
    }
	//
	SS80_talking = 0;
	SS80_listening = 0;
	spoll = 0;
	SS80_tstate = SS80_TRAN_IDLE;
	SS80_cstate = SS80_START;
	qstat = 2;							// 2 Indicates that the device has just been powered up.
	
	EnablePP();
}
//
// This is the transaction and command parser. Parses byte by byte from gpib. Return a command token to stop GPIB reception and process a command
//
// Parsing is governed by two state-variables: cstate and tstate. tstate is the transaction state and is also used for non-transaction commands (transparent)
// cstate is the actual command state that handles reading of commands with parameters and also persists the command from command state to
// execution state
//
int SS80_char(unsigned char ch, unsigned char atn, unsigned char reset)
{
unsigned char status;
//#if SDEBUG > 1
//if(SS80_listening)
//    xprintf(" {L%u %u %d}[%02X]", SS80_cstate,SS80_tstate,atn,ch);
//else if(SS80_talking)
//    xprintf(" {T%u %u %d}[%02X]", SS80_cstate,SS80_tstate,atn,ch);
//else
//    xprintf(" {X%u %u %d}[%02X]", SS80_cstate,SS80_tstate,atn,ch);
//#endif
	if(reset) SS80_init();
	status = 0;
	//
	// Handle single-byte commands, they all reset the state-machine in various ways
	//
	if(atn)
	{
#if defined(SPOLL)
		if(ch == SPE)									// If we see an serial oll enable (SPE) then set
		{												// flag. If we happen to be talking, then send
			spoll = 1;									// an spoll response. Apparently does the 44788 to it
			SS80_tstate = SS80_TRAN_IDLE;				// this way around.
			if(SS80_talking)
			{
				gpib_tx_buff[0] = qstat;
				sendgpib(1);
			}
			return 0;
		}
		if(ch == SPD)
		{
			spoll = 0;
			SS80_tstate = SS80_TRAN_IDLE;
			return 0;
		}
#endif
		if(ch == SDC || ch == DCL)						// Selected device clear or all device clear
		{
			status = SS80_CMD_CLEAR;
			SS80_tstate = SS80_TRAN_IDLE;
			SS80_cstate = SS80_START;
			return status;
		}
		if(ch == UNL)
		{
			SS80_listening = 0;
			SS80_tstate = SS80_TRAN_IDLE;
			return 0;
		}
		if(ch == UNT)
		{
			SS80_talking = 0;
			SS80_tstate = SS80_TRAN_IDLE;
			return 0;
		}
		if(ch >= 0x40 && ch <= 0x5e && ch != MTA)		// A talk address automatically untalks everyone
		{
			SS80_talking = 0;
			SS80_tstate = SS80_TRAN_IDLE;
			return 0;
		}
		if(ch == MTA)
		{
			SS80_talking = 1;
			SS80_tstate = SS80_TRAN_IDLE;
#if SDEBUG > 1
			xprintf("[MTA]");
#endif

#if defined(SPOLL)
			if(spoll)									// If spoll has been enabled (SPE) and we are
			{											// set to talk, then send an spoll response
				gpib_tx_buff[0] = qstat;
				sendgpib(1,0);
			}
#endif
			return 0;
		}
		if(ch == MLA)
		{
			SS80_listening = 1;
			SS80_tstate = SS80_TRAN_IDLE;
//#if SDEBUG > 1
			xprintf("[MLA]");
//#endif
			return 0;
		}
		// ---------------------------------------------------
		if(ch == 0x65 && SS80_listening)		// Go to command state
		{
			SS80_cstate = SS80_START;
			SS80_tstate = SS80_TRAN_COMMAND;
			return 0;
		}
		if(ch == 0x6E)		// Go to Execution state (both listen & talk)
		{
			SS80_tstate = SS80_TRAN_EXECUTION;
		}
		if(ch == 0x70 && SS80_talking)		// Go to report state
		{
			SS80_cstate = SS80_START;
			SS80_tstate = SS80_TRAN_REPORT;
		}
		if(ch == 0x70 && SS80_listening)		// This is Amigo clear
		{
			DisablePP();
			SS80_cstate = SS80_START;
			SS80_tstate = SS80_TRAN_AMIGO_CLEAR;
			return 0;					// Return and wait for next byte
		}
		if(ch == 0x72 && SS80_listening)		// This is transparent command
		{
			DisablePP();
			SS80_cstate = SS80_072_L;
			SS80_tstate = SS80_TRANSPARENT;
			return 0;					// Return and wait for next byte
		}
	}
	//
	//if(ch!=0) xprintf("\r\ntstate %d, ch %X", SS80_tstate, ch);
	//
	switch(SS80_tstate)
	{
		case SS80_TRAN_IDLE:
			//
			break;
		case SS80_TRAN_COMMAND:
			status = SS80_do_command(ch);
			break;
		case SS80_TRAN_EXECUTION:
			DisablePP();
			status = SS80_exec_command(ch);
			SS80_tstate = SS80_TRAN_IDLE;	//Set idle after cmd to shorten the time for arriving bytes...
			break;
		case SS80_TRAN_REPORT:				// Report status
			DisablePP();					// Disable parallell poll
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Rpt QSTAT %d ", TMR1H, TMR1L, qstat);
#endif
			memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
			gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on
			gpib_tx_buff[0] = qstat;
            //
            // this is a bit odd. The HP1631D needs about this much delay here
            //
            xprintf("\r\nTXDATA len %d ", 1);
            DumpData(gpib_tx_buff, 1, "TXData");
            
			sendgpib(1);
			gpib_flags = gpib_flags & ~FLAGS_EOI;	// EOI off
			status = SS80_CMD_NOP;
            SS80_tstate = SS80_TRAN_IDLE; // THIS!
			break;

		case SS80_TRAN_AMIGO_CLEAR:			// Status byte from amigo clear. 0=Disable parity, 1=Enable parity
			EnablePP();
			status = 0;
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Amigo Clear %d ", TMR1H, TMR1L, ch);
#endif
			break;
											// TODO: Look at the other transparent ones: 01, 02, 03, 09
		case SS80_TRANSPARENT:
			if(ch == 0x08)					// 0x08 is Channel Independent clear
			{
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Ch ind Clear %d ", TMR1H, TMR1L, ch);
#endif
				EnablePP();	
				status = 0;
				qstat = 0;					// Clear clears qstat
				SS80_tstate = SS80_TRAN_IDLE;
				SS80_cstate = SS80_START;
				break;
			}
		default:
			break;
	}
	//xprintf("(%d)", status);
	return status;
}
//
// Act on a SS/80 command
//
BYTE SS80_do_command(BYTE ch)
{
BYTE status = 0;

	switch(SS80_cstate)
	{
		case SS80_START:
			if(ch == 0x00 && SS80_listening)		// Read
			{
#if SDEBUG >= 1
				xprintf("\r\n%02X%02X SS80:Read ", TMR1H, TMR1L);
#endif
				status = 0;
				SS80_cstate = SS80_000;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch == 0x02 && SS80_listening)		// Write
			{
#if SDEBUG >= 1
				xprintf("\r\n%02X%02X SS80:Write ",TMR1H, TMR1L);
#endif
				status = 0;
				SS80_cstate = SS80_002;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch >= 0x20 && ch <= 0x2f && SS80_listening)		// Set unit
			{
				Unit = ch - 0x20;
#if SDEBUG >= 1
				xprintf("\r\n%02X%02X SS80:Set unit %d ", TMR1H, TMR1L, Unit);
#endif
				status = 0;
                //
                // Does the image file exist? If not signal error
                //
                if(File[Unit].fsize < 1 || Unit > MAXDISK)
                {
                    qstat = 1;
                    SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;
                }
				SS80_cstate = SS80_START;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch >= 0x40 && ch <= 0x4f && SS80_listening)		// Set unit
			{
				SS80Disk[Unit].Volume = ch - 0x40;
#if SDEBUG >= 1
				xprintf("\r\n%02X%02X SS80:Set volume %d ", TMR1H, TMR1L, SS80Disk[Unit].Volume);
#endif
				status = 0;
				SS80_cstate = SS80_START;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch == 0x34 && SS80_listening)		// No operation
			{
#if SDEBUG >= 1
				xprintf("\r\n%02X%02X SS80:NOP ", TMR1H, TMR1L);
#endif
				SS80_cstate = SS80_START;
				status = 0;
                EnablePP();
				break;
			}
			if(ch == 0x35 && SS80_listening)		// Describe
			{
#if SDEBUG >= 1
				xprintf("\r\n%02X%02X SS80:Describe ", TMR1H, TMR1L);
#endif
				status = 0;
				SS80_cstate = SS80_035;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch == 0x37 && SS80_listening)		// Initialise media, 2 bytes follow
			{
				SS80_cstate = SS80_037_1;
				break;
			}
			if(ch == 0x0d && SS80_listening)		// Request status
			{
				status = 0; // SS80_CMD_SEND_STATUS;
				SS80_cstate = SS80_00D;
				EnablePP();
#if SDEBUG >= 1
				xprintf("\r\n%02X%02X SS80:Rq status ", TMR1H, TMR1L);
#endif
				break;
			}
			if(ch == 0x10 && SS80_listening)		// Set address, 6 bytes follow
			{
				SS80_cstate = SS80_010_1;
				break;
			}
			if(ch == 0x18 && SS80_listening)		// Set length, 4 bytes follow
			{
				SS80_cstate = SS80_018_1;
				break;
			}
			if(ch == 0x3E && SS80_listening)		// Set status mask, 8 bytes follow
			{
				SS80_cstate = SS80_03E_1;
				break;
			}
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Unknown %X ", TMR1H, TMR1L, ch);
#endif
			break;

		case SS80_010_1:
			SS80Disk[Unit].Address[0] = ch;	// MSB
			SS80_cstate = SS80_010_2;
			break;
		case SS80_010_2:
			SS80Disk[Unit].Address[1] = ch;
			SS80_cstate = SS80_010_3;
			break;
		case SS80_010_3:
			SS80Disk[Unit].Address[2] = ch;
			SS80_cstate = SS80_010_4;
			break;
		case SS80_010_4:
			SS80Disk[Unit].Address[3] = ch;
			SS80_cstate = SS80_010_5;
			break;
		case SS80_010_5:
			SS80Disk[Unit].Address[4] = ch;
			SS80_cstate = SS80_010_6;
			break;
		case SS80_010_6:					// TODO. Only handles 4-byte vectors, 0xffffffff bytes = 4 294 967 295 bytes

											// What we should do though is a union
											// Address 00 00 00 00 00 00
											// Vector           00 00 00 00
											//
											// typedef union {
											//  DWORD Vector;
											//  BYTE Addressbytes[4];
											// } AddressType;
											//
											// ie we grab the three LSB bytes from the address and "multiply" by 100h when stuffing into the union
			SS80Disk[Unit].Address[5] = ch;
			SS80Disk[Unit].Vector = SS80Disk[Unit].Address[2] * 0x100000000L + SS80Disk[Unit].Address[3] 
                    * 0x1000000L + SS80Disk[Unit].Address[4] * 0x10000L + SS80Disk[Unit].Address[5] * 0x100L;
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Set address %lX ",TMR1H, TMR1L, SS80Disk[Unit].Vector);
#endif
			SS80_cstate = SS80_START;	// and back to the top for next cmd
			break;

		case SS80_018_1:
			SS80Disk[Unit].Length.Lenbytes[3] = ch;	// MSB
			SS80_cstate = SS80_018_2;
			break;
		case SS80_018_2:
			SS80Disk[Unit].Length.Lenbytes[2] = ch;
			SS80_cstate = SS80_018_3;
			break;
		case SS80_018_3:
			SS80Disk[Unit].Length.Lenbytes[1] = ch;
			SS80_cstate = SS80_018_4;
			break;
		case SS80_018_4:
			SS80Disk[Unit].Length.Lenbytes[0] = ch;	// LSB
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Set length %lX ",TMR1H, TMR1L, SS80Disk[Unit].Length.Length);
#endif
			SS80_cstate = SS80_START;
			break;

        case SS80_03E_1:                // 8 bytes of status mask
            StatusMask[0] = ch;
			SS80_cstate = SS80_03E_2;
			break;
        case SS80_03E_2:
            StatusMask[1] = ch;
			SS80_cstate = SS80_03E_3;
			break;    
        case SS80_03E_3:
            StatusMask[2] = ch;
			SS80_cstate = SS80_03E_4;
			break;   
        case SS80_03E_4:
            StatusMask[3] = ch;
			SS80_cstate = SS80_03E_5;
			break;  
        case SS80_03E_5:
            StatusMask[4] = ch;
			SS80_cstate = SS80_03E_6;
			break;   
        case SS80_03E_6:
            StatusMask[5] = ch;
			SS80_cstate = SS80_03E_7;
			break;  
        case SS80_03E_7:
            StatusMask[6] = ch;
			SS80_cstate = SS80_03E_8;
			break; 
        case SS80_03E_8: 
            StatusMask[7] = ch;
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Set Status mask %X %X %X %X %X %X %X %X ",TMR1H, TMR1L,
                StatusMask[0],
                StatusMask[1],
                StatusMask[2],
                StatusMask[3],
                StatusMask[4],
                StatusMask[5],
                StatusMask[6],
                StatusMask[7]
                );
#endif
			SS80_cstate = SS80_START;
			break;
		case SS80_034:					// No operation	
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:NOP ", TMR1H, TMR1L);
#endif
			SS80_cstate = SS80_START;		
			break;

		case SS80_035:					
			break;
										// Initialise disk command
		case SS80_037_1:				// First byte of init params
			SS80_cstate = SS80_037_2;
			break;
		case SS80_037_2:				// Second byte of init param
#if SDEBUG >= 1
			xprintf("\r\n%02X%02X SS80:Init disk ", TMR1H, TMR1L);
#endif
			SS80_cstate = SS80_START;
			status = SS80_CMD_FORMAT;
			EnablePP();
			break;
		case SS80_072_L:				// Transparent commands beginning with 072

		case SS80_END:
			break;
	}
	return status;
}
//
// Execute a SS/80 command
//
BYTE SS80_exec_command(BYTE ch)
{
BYTE status;

	switch(SS80_cstate)
	{
		case SS80_000:
			status = SS80_CMD_BUFFERED_READ;
			break;
		case SS80_002:
			status = SS80_CMD_BUFFERED_WRITE;
			SS80_cstate = SS80_START;
			break;
		case SS80_035:
			status = SS80_CMD_DESCRIBE;
			break;
		case SS80_00D:
			status = SS80_CMD_SEND_STATUS;
			break;
		case SS80_END:
			break;
		default:
			status = 0;		// Must return zero if we do nothing!!!
			break;
	}
	return status;
}
//
// Return drive parameters to host
//
void SS80_cmd_describe()
{
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80x:Describe ", TMR1H, TMR1L);
#endif
	//
	gpib_flags = gpib_flags & ~FLAGS_EOI;	// No EOI
	memcpy(gpib_tx_buff,(void *)&ControllerDescription,5);
	//
    memcpy(&gpib_tx_buff[5],(void*)&UnitDescription,19);
	//
	gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on
	memcpy(&gpib_tx_buff[24],(void *)&VolumeDescription.V1,13);
    sendgpib(37);
	//
	gpib_flags = gpib_flags & ~FLAGS_EOI;
    //trigger(5);
	EnablePP();
}
//
// Return Status to host
//
// Only status we bother returning for now is write protect, unit fault and
// address bounds
//
void SS80_cmd_send_status()
{
    int i;
    BYTE mask;
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80x:Status, unit %d ", TMR1H, TMR1L, Unit);
#endif
	//
	gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	gpib_tx_buff[0] = Unit; //0x11;                 // Bytes 1 = Volume 1, Unit 1
	gpib_tx_buff[1] = 0xff;
	//
    // Bytes 3-10 (2..9) are the error status bytes
    //
	if(SS80Disk[Unit].Errors & SS80_ERR_SEEK) gpib_tx_buff[4] = 0b00000001; // address bounds = Byte 3
	if(SS80Disk[Unit].Errors & SS80_ERR_READ) gpib_tx_buff[6] = 0b00000010; // unit fault = Byte 5
	if(SS80Disk[Unit].Errors & SS80_ERR_WRITE) gpib_tx_buff[6] = 0b00000010; // unit fault = Byte 5
	if(WP_PORT) gpib_tx_buff[6] = 0b00001000; // Write protect = Byte 7
	//
    // But they can be masked by a devious Set Status Mask...
    // an 1 in any position masks the error bit so invert and AND
    //
    for(i=0;i<7;i++)
    {
        xprintf("\r\n%02X%02X SS80x:Status[%d] %X ", TMR1H, TMR1L, i, gpib_tx_buff[i+2]);
        mask = ~StatusMask[i];
        gpib_tx_buff[i+2] = gpib_tx_buff[i+2] & mask;
        xprintf("-> %X ", gpib_tx_buff[i+2]);
#if SDEBUG >= 1
	
#endif
    }
	qstat = SS80Disk[Unit].Errors;
	//
	gpib_tx_buff[11] = SS80Disk[Unit].Address[0];		// Byte 11-16 = target address
	gpib_tx_buff[12] = SS80Disk[Unit].Address[1];
	gpib_tx_buff[13] = SS80Disk[Unit].Address[2];
	gpib_tx_buff[14] = SS80Disk[Unit].Address[3];
	gpib_tx_buff[15] = SS80Disk[Unit].Address[4];
	gpib_tx_buff[16] = SS80Disk[Unit].Address[5];
	sendgpib(20);

	gpib_flags = gpib_flags & ~FLAGS_EOI;
	EnablePP();
}
//
// Seek to the position in vector (called by read and write). Returns nonzero on error
//
int SS80_cmd_seek()
{
DWORD fpos;
FRESULT rc;
	fpos = SS80Disk[Unit].Vector;
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80x:Seek: Seeking to %lu (%lX) ", TMR1H, TMR1L, fpos, fpos);
#endif
	//
	// Bounds checking
	//
	if (SS80Disk[Unit].Vector > SS80Disk[Unit].Maxvector)
	{
#if SDEBUG >= 1
		xprintf("\r\nSeek beyond end");
#endif
		qstat = 1;
		SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;		
	}
	if ((SS80Disk[Unit].Vector + SS80Disk[Unit].Length.Length) > SS80Disk[Unit].Maxvector)
	{
#if SDEBUG >= 1
		xprintf("\r\nRead beyond end");
#endif
		qstat = 1;
		SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;		
	}
	qstat = 0;
    //
	rc=f_lseek(&File[Unit],fpos);
	if(rc)
	{
#if SDEBUG >= 1
		xprintf("\r\nSeek to %ld failed %d", fpos, rc);
#endif
		qstat = 1;
		SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;
	}
	return qstat;
}
//
// Do a buffered read of the current sector
//
// Data is (for now) always sent in 256-bit chunks, because of how the GPIB TX buffer was once set up. So for now
// we calculate how many 256-bit chunks to send and then send the rest. Return nonzero on error
//
int SS80_cmd_buffered_read()
{
int i,rc2;
UINT total_bytes;
FRESULT rc;
UINT nbytes;
DS2417 t;
int chunks;				// Number of 256 byte chunks to send
int tail;				// tail after the chunks
						// Send Length bytes. EOI on last packet

	DisablePP();
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80x:Read %lu (%lX) ", TMR1H, TMR1L, SS80Disk[Unit].Vector, SS80Disk[Unit].Vector);
#endif
	rc2 = SS80_cmd_seek();
	if(rc2 != 0)
	{
		SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;
		SS80_error_return();		// return one byte and
		return rc2;					// bail out if seek fails
	}
    //
    // Reading the RTC file??
    // Data is four integers since the 98x6 has no unsigned integers, with
    // the actual byte in the lower part.
    //
    if(SS80Disk[Unit].Vector == (MagicVector+1)*256)
    {
        t = (DS2417) DS2417_read();
        gpib_tx_buff[0]=0;
        gpib_tx_buff[1]=t.ch[3];
        gpib_tx_buff[2]=0;
        gpib_tx_buff[3]=t.ch[2];
        gpib_tx_buff[4]=0;
        gpib_tx_buff[5]=t.ch[1];
        gpib_tx_buff[6]=0;
        gpib_tx_buff[7]=t.ch[0];
        gpib_flags = gpib_flags & ~FLAGS_EOI;
        sendgpib(256);
        gpib_flags = gpib_flags & ~FLAGS_EOI;
#if SDEBUG >= 1
        xprintf("\r\n%02X%02X Sending RTC bytes %2X %2X %2X %2X ", TMR1H, TMR1L, gpib_tx_buff[1],gpib_tx_buff[3],gpib_tx_buff[5],gpib_tx_buff[7]);
#endif
        qstat = 0;
        EnablePP();
        return qstat;  
    }
	//
	//								// Calculate how to send our data
	chunks = SS80Disk[Unit].Length.Length / 256;	// Whole 256-byte blocks
	tail = SS80Disk[Unit].Length.Length % 256;		// and possible tail < 256
	//
	gpib_flags = gpib_flags & ~FLAGS_EOI;
	total_bytes = 0;

	for(i=0;i<chunks;i++)	// Read maximum 256 bytes
	{
		if(i==(chunks-1) && tail == 0) gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on, on last block if there is no tail
//		xprintf("Rd to: %d ", gpib_tx_buff);
        rc=f_read(&File[Unit],gpib_tx_buff,256,&nbytes);
#if SDEBUG >= 1
		xprintf("\r\n%02X%02X 1:Read rtn %X, %d read ", TMR1H, TMR1L, rc, nbytes);
#endif
		if(rc != 0)
		{
			SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;
			SS80_error_return();		// return one byte and
			return rc2;					// bail out if seek fails
		}
		//Time();
		sendgpib(nbytes);
		total_bytes = total_bytes + nbytes;
		SS80Disk[Unit].Vector = SS80Disk[Unit].Vector + 0x100L;
#if SDEBUG >= 1
		xprintf("\r\n%02X%02X Vector %ld (%lX) ", TMR1H, TMR1L, SS80Disk[Unit].Vector, SS80Disk[Unit].Vector);
#endif
	}
	// Send the rest, in case it did not fit in the chunks
	if(tail > 0)
	{
		rc=f_read(&File[Unit],gpib_tx_buff,tail,&nbytes);
#if SDEBUG >= 1
		xprintf("\r\n2:Read rtn %X, %d read ", rc, nbytes);
#endif
		if(rc != 0)
		{
			SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;
			SS80_error_return();		// return one byte and
			return rc2;					// bail out if read fails
		}
		gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on, on tail block
		sendgpib(tail);
		total_bytes = total_bytes + nbytes;
		SS80Disk[Unit].Vector = SS80Disk[Unit].Vector + 0x100L;
#if SDEBUG >= 1
		xprintf("\r\nVector %ld (%lX) ", SS80Disk[Unit].Vector, SS80Disk[Unit].Vector);
#endif
	}
	gpib_flags = gpib_flags & ~FLAGS_EOI;
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X %d bytes sent ", TMR1H, TMR1L, total_bytes);
#endif
	qstat = 0;
	EnablePP();
	return qstat;
}
//
// Receive data and call write_sector() for each chunk of 256 bytes. returns nonzero on error
//
int SS80_cmd_buffered_write()
{
int more;
int gpib_bytes;
DWORD total_bytes;
FRESULT rc;
DS2417 t;
int rc2;		// readgpib() returns an integer, silly!
char b,c;

    DisablePP();
	qstat = 0;
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80x:Write %ld (%lX) ", TMR1H, TMR1L, SS80Disk[Unit].Vector, SS80Disk[Unit].Vector);
#endif
	rc2 = SS80_cmd_seek();
	if(rc2 != 0) return rc2;		// Bail out if seek fails
	//
	total_bytes = 0;
	more = 1;
    c = 0;
	while(more)							// Loop until we have received Length sectors or EOI
	{
        gpib_bytes = 0;
        while(gpib_bytes < 256) // && timeout < 10000)
        {
            if(gpib_test())             // We have a character from GPIB
            {
                b = gpib_getc(&c);
                //xprintf(" %X %X %d ", b, c, gpib_bytes);
                gpib_rx_buff[gpib_ptr++] = b;
                ++gpib_bytes;
#if SDEBUG >= 9
                xprintf("\r\n %02X %c %4s %4s %4s %4s %4s %d",
                b,b,
                c&IFC_BIT ? "IFC ":"  ",
                c&SRQ_BIT ? "SRQ ":"  ",
                c&ATN_BIT ? "ATN ":"  ",
                c&EOI_BIT ? "EOI ":"  ",
                c&REN_BIT ? "REN ":"  ",
                        gpib_ptr);
#endif
                //if(c&EOI_BIT) break;
                
			// IFC
            }
            if(c&EOI_BIT) break;
            if(IFC_PORT == 0) break;
        }

        total_bytes = total_bytes + gpib_bytes;
        if(total_bytes >= SS80Disk[Unit].Length.Length) more = 0;	// All done?
        if(c&EOI_BIT) more = 0; 						// EOI		
        if(IFC_PORT == 0) more = 0;		
#if SDEBUG >= 1
		if(gpib_ptr > 0)
		{
			xprintf("\r\nGPIB rc:%d %d bytes\r\n", rc2,gpib_ptr);
		}
#endif
		//	-----------------------------------

		//
		// SS/80 behaviour: Replicate last byte in case it is a partial write
		//
		if(gpib_bytes < 256)
		{
			memset(&gpib_rx_buff[gpib_ptr],gpib_rx_buff[gpib_ptr-1],256-gpib_ptr);
		}
        //
        // Writing the RTC file??
        //
        if (SS80Disk[Unit].Vector == (MagicVector + 1)*256)
        {
            t.ch[3] = gpib_rx_buff[1];
            t.ch[2] = gpib_rx_buff[3];
            t.ch[1] = gpib_rx_buff[5];
            t.ch[0] = gpib_rx_buff[7];
            DS2417_init(t.time);
#if SDEBUG >= 1
            xprintf("\r\n%02X%02X Received RTC bytes %2X %2X %2X %2X ", TMR1H, TMR1L, gpib_tx_buff[0], gpib_tx_buff[1], gpib_tx_buff[2], gpib_tx_buff[3]);
#endif
        }
		rc=Write_sector();
		if(rc)
		{
#if SDEBUG >= 1
			xprintf("\r\nWrite_sectors fail %d", rc);
#endif
			qstat = 1;
			SS80Disk[Unit].Errors = SS80Disk[Unit].Errors | SS80_ERR_SEEK;
			return qstat;
		}
	}
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80:Buff wr %ld bytes rx ", TMR1H, TMR1L, total_bytes);
#endif
							// The address is incremented on a successful write by Write_sectors().
	if(WP_PORT) qstat = 1;
	EnablePP();
	return 0;
}
//
// QSTAT, return status byte
//
void SS80_cmd_qstat()
{
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80:DSJ ", TMR1H, TMR1L);
#endif
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	gpib_tx_buff[0] = qstat;
	sendgpib(1);	
	//
	// QSTAT clears the qstat byte that might be 2 after powerup
	//
	qstat = 0;
}
//
// Response to selected device clear or all clear - Select sector 0
//
//
void SS80_cmd_clear()
{
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80x:Clear ", TMR1H, TMR1L);
#endif
    for(int i = 0; i<=MAXDISK;i++)
    {
        SS80Disk[Unit].Vector = 0;
        SS80Disk[Unit].Errors = 0;
        SS80Disk[Unit].Length.Length = 0;
    }
	EnablePP();
	//
	// Clear the DSJ byte that might be 2 after powerup
	//
	qstat = 0;
}
//
// Increment vector
//
void SS80_increment()
{
UINT fpos;
FRESULT rc;

	SS80Disk[Unit].Vector = SS80Disk[Unit].Vector + 0x100L;
	fpos = SS80Disk[Unit].Vector;
#if SDEBUG >= 1
	xprintf("\r\n%02X%02X SS80: Increment: Seeking to %lu (%lX) ", TMR1H, TMR1L, fpos, fpos);
#endif
	rc=f_lseek(&File[Unit],fpos);
	if(rc)
	{
#if SDEBUG >= 1
		xprintf("\r\nSeek to %ld failed %d", fpos, rc);
#endif
		qstat = 1;
	}
	else qstat = 0;
}

int Write_sector()
{
// Write 256 bytes in gpib_rx_buff at current position, then advance to next sector
//
// Returns 0 on success, nonzero otherwise
//
int rc2;
UINT sector;
UINT nbytes;
FRESULT rc;
int status;

	sector = SS80Disk[Unit].Vector / 0x100L;
	status = 0;
#if SDEBUG > 1
	xprintf("\r\nWrite %ld (%lX) s:%lU ", SS80Disk[Unit].Vector, SS80Disk[Unit].Vector,sector);
	DumpData(gpib_rx_buff, 256,"RXbuf");
#endif
	rc2 = SS80_cmd_seek();
	if(rc2 != 0) return rc2;		// Bail out if seek fails
	//
	if(!WP_PORT) rc=f_write(&File[Unit],(void*)gpib_rx_buff, 256, &nbytes);	// Always write 256 bytes, ie a full block
	if(rc)
	{
#if SDEBUG >= 1
		xprintf("\r\nWrite fail %d", rc);
#endif
		status = 1;
		return rc;
	}

	if(status == 0)
    {
        SS80_increment();
        writeflag = 1;      //Indicate we have data to be synced;
    }
	return status;
}
//
// Error return. On catastrophic failures, just return one byte, Value 1, with EOI
//
void SS80_error_return()
{
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	gpib_tx_buff[0] = 1;
	gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on
	sendgpib(1);	
	gpib_flags = gpib_flags & ~FLAGS_EOI;
}
//------------------------- Test routines --------------------
//char TD[] = {0x3f,1,0x41,1,0x20,1,0x65,1,
//0x35,0,
//0x3f,1,
//0x3f,1,
//0x21,1,
//0x40,1,
//0x6e,1, // exec phase secondary
//0x5f,1,
//0x3f,1,
//0x21,1,
//0x40,1,
//0x70,1,	// Rpt phase secondary
//0x3F,1, // UNL
//0x5F,1	// UNT
//};
//
// Test code below, comment out
//char TD[] = {
//0x3F,0,
//0x5f,1,
//0x20,1,
//0x65,1,
//0x20,0,
//0x40,0, 
//0x34,0, // NOP
//0x37,0, // Format
//0x00,0,
//0x0a,0
//};
//
// Format a disk. Rename the old lifdata file and create a new. Write all blocks
//
void SS80_cmd_format()
{
UINT i;
UINT nbytes;
BYTE res;

    DisablePP();
    qstat = 1;                      // Assume error
//    res = f_close(&File[Unit]);
//    if(res != FR_OK)
//    {
//        sxprintf(string, "fclose returns %d ",res);
//        sio_puts(string);
//        return;
//    }
//    res = f_unlink("lifdata.old");
//    if(res != FR_OK)
//    {
//        sxprintf(string, "unlink returns %d ",res);
//        sio_puts(string);
//        return;
//    }
//    res = f_rename(filename, "lifdata.old");
//    if(res != FR_OK)
//    {
//        sxprintf(string,"rename returns %d ",res);
//        sio_puts(string);
//        return;
//    }
//    //
//    sxprintf(string, "\nOpening new %s ", filename);
//    sio_puts(string);
//    res = f_open(&File[Unit], filename, FA_CREATE_NEW + FA_WRITE + FA_READ);
//    if(res != FR_OK)
//    {
//        sxprintf(string, "fopen returns %d ",res);
//        sio_puts(string);
//        return;
//    }
//    //
//    sxprintf(string, "Formatting %ld (%lX)", Maxvector, Maxvector);
//    sio_puts(string);
//    //xprintf("\r\nWrite %ld (%lX) s:%ld ", Vector, Vector,sector);
//    for(i=0;i<Maxvector/0x100L;i++)
//    //for(i=0;i<100;i++)
//	{
//        sxprintf(string, "block %ul ",i);
//        sio_puts(string);
//		if(!WP_PORT) res=f_write(&File[Unit],(void*)gpib_rx_buff, 256, &nbytes);
//        if(res != FR_OK)
//        {
//            sxprintf(string, "rc=%d ",res);
//            sio_puts(string);
//            return;
//        }
//	}
//    sxprintf(string, "\nSyncing %s ", filename);
//    sio_puts(string);
//    res = f_sync(&File[Unit]);
//    if(res != FR_OK)
//    {
//        sxprintf(string, "fsync returns %d ",res);
//        sio_puts(string);
//        return;
//    }
    qstat = 0;                      // No error
	EnablePP();
}

//void SS80_Test()
//{
//DWORD i;
//BYTE rc;
//
//for(i=0;i<(Vector/0x100L);i++)
//	{
//    
//		rc=SS80_char(TD[i],TD[i+1],0);
//		xprintf("\r\nrc=%d",rc);
//	}
//}
